import pandas as pd
import pytest

from sdgx.data_models.inspectors.i_id import IDInspector


@pytest.fixture
def inspector():
    yield IDInspector()


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


@pytest.fixture
def id_test_df():
    row_cnt = 1000
    header = ["int_id", "str_id", "not_int_id", "not_str_id"]

    int_id = list(range(row_cnt))
    str_id = list("id_" + str(i) for i in range(row_cnt))

    not_int_id = list(range(int(row_cnt / 2))) + list(range(int(row_cnt / 2)))
    not_str_id = list("id_" + str(i) for i in range(int(row_cnt / 2))) + list(
        "id_" + str(i) for i in range(int(row_cnt / 2))
    )

    X = [[int_id[i], str_id[i], not_int_id[i], not_str_id[i]] for i in range(row_cnt)]

    yield pd.DataFrame(X, columns=header)


def test_inspector_demo_data(inspector: IDInspector, raw_data):
    inspector.fit(raw_data)
    assert inspector.ready
    # should be empty set
    assert not inspector.ID_columns
    assert sorted(inspector.inspect()["id_columns"]) == sorted([])
    assert inspector.inspect_level == 20


def test_inspector_generated_data(inspector: IDInspector, id_test_df: pd.DataFrame):
    # use generated id data
    inspector.fit(id_test_df)
    assert inspector.ID_columns
    assert sorted(inspector.inspect()["id_columns"]) == sorted(["int_id", "str_id"])


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from pydantic import BaseModel

from sdgx.exceptions import SdgxError


class ExitMessage(BaseModel):
    code: int
    msg: str
    payload: dict = {}

    def _dump_json(self) -> str:
        return self.model_dump_json()

    def send(self):
        print(self._dump_json(), flush=True, end="")


class NormalMessage(ExitMessage):
    code: int = 0
    msg: str = "Success"

    @classmethod
    def from_return_val(cls, return_val) -> "NormalMessage":
        if isinstance(return_val, dict):
            payload = return_val
        else:
            payload = {"return_val": return_val}
        return cls(code=0, msg="Success", payload=payload)


class ExceptionMessage(ExitMessage):
    @classmethod
    def from_exception(cls, e: Exception) -> "ExceptionMessage":
        if isinstance(e, SdgxError):
            return cls(
                code=e.ERROR_CODE,
                msg=str(e),
                payload={
                    "details": "Synthetic Data Generator Error, please check logs and raise an issue on https://github.com/hitsz-ids/synthetic-data-generator."
                },
            )
        return cls(
            code=-1,
            msg=str(e),
            payload={
                "details": "Unknown Exceptions, please check logs and raise an issue on https://github.com/hitsz-ids/synthetic-data-generator."
            },
        )



"""Univariate selection function."""

import numpy as np
from scipy.stats import kstest

from sdgx.models.components.sdv_copulas import get_instance


def select_univariate(X, candidates):
    """Select the best univariate class for this data.

    Args:
        X (pandas.DataFrame):
            Data for which be best univariate must be found.
        candidates (list[Univariate]):
            List of Univariate subclasses (or instances of those) to choose from.

    Returns:
        Univariate:
            Instance of the selected candidate.
    """
    best_ks = np.inf
    best_model = None
    for model in candidates:
        try:
            instance = get_instance(model)
            instance.fit(X)
            ks, _ = kstest(X, instance.cdf)
            if ks < best_ks:
                best_ks = ks
                best_model = model
        except Exception:
            # Distribution not supported
            pass

    return get_instance(best_model)



"""Independence module."""

import numpy as np

from sdgx.models.components.sdv_copulas.bivariate.base import Bivariate, CopulaTypes
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix


class Independence(Bivariate):
    """This class represent the copula for two independent variables."""

    copula_type = CopulaTypes.INDEPENDENCE

    def fit(self, X):
        """Fit the copula to the given data.

        Args:
            X (numpy.array): Probabilites in a matrix shaped (n, 2)

        Returns:
            None

        """

    def generator(self, t):
        """Compute the generator function for the Copula.

        The generator function is a function f(t), such that an archimedian copula can be
        defined as

        C(u1, ..., uN) = f(f^-1(u1), ..., f^-1(uN)).

        Args:
            t(numpy.array)

        Returns:
            np.array

        """
        return np.log(t)

    def probability_density(self, X):
        """Compute the probability density for the independence copula."""
        return np.all((0.0 <= X) & (X <= 1.0), axis=1).astype(float)

    def cumulative_distribution(self, X):
        """Compute the cumulative distribution of the independence bivariate copula is the product.

        Args:
            X(numpy.array): Matrix of shape (n,2), whose values are in [0, 1]

        Returns:
            numpy.array: Cumulative distribution values of given input.

        """
        U, V = split_matrix(X)
        return U * V

    def partial_derivative(self, X):
        """Compute the conditional probability of one event conditiones to the other.

        In the case of the independence copula, due to C(u,v) = u*v, we have that
        F(u|v) = dC/du = v.

        Args:
            X()

        """
        _, V = split_matrix(X)
        return V

    def percent_point(self, y, V):
        """Compute the inverse of conditional cumulative distribution :math:`F(u|v)^-1`.

        Args:
            y: `np.ndarray` value of :math:`F(u|v)`.
            v: `np.ndarray` given value of v.

        """
        self.check_fit()
        return y



import pandas as pd
import pytest

from sdgx.data_models.inspectors.manager import InspectorManager
from sdgx.exceptions import InspectorInitError


@pytest.fixture
def manager():
    yield InspectorManager()


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


@pytest.mark.parametrize(
    "basic_inspector",
    ["DiscreteInspector", "NumericInspector", "IDInspector", "BoolInspector", "DatetimeInspector"],
)
def test_manager(basic_inspector, manager: InspectorManager):
    assert manager._normalize_name(basic_inspector) in manager.registed_inspectors


@pytest.mark.parametrize("inspector_name", list(InspectorManager().registed_inspectors.keys()))
def test_inspector(inspector_name, manager: InspectorManager, raw_data: pd.DataFrame):
    each_inspector = manager.init(inspector_name)
    # check type
    assert "sdgx.data_models.inspectors" in str(type(each_inspector))
    # check ready
    assert each_inspector.ready is False
    if not "relationship" in inspector_name:
        each_inspector.fit(raw_data)
        assert each_inspector.ready is True
    # check inspect level property
    assert each_inspector.inspect_level <= 100 and each_inspector.inspect_level > 0
    # check inspect_level.setter
    # set level
    each_inspector.inspect_level = 66
    assert each_inspector.inspect_level == 66
    each_inspector.inspect_level = 88
    assert each_inspector.inspect_level == 88
    each_inspector.inspect_level = 100
    assert each_inspector.inspect_level == 100
    each_inspector.inspect_level = 10
    assert each_inspector.inspect_level == 10
    # test error 1
    has_error = False
    try:
        each_inspector.inspect_level = 101
    except Exception as e:
        has_error = True
        assert type(e) == InspectorInitError
    assert has_error is True
    # test error 2
    has_error = False
    try:
        each_inspector.inspect_level = 0
    except Exception as e:
        has_error = True
        assert type(e) == InspectorInitError
    assert has_error is True


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.transformers.numeric import NumericValueTransformer


@pytest.fixture
def df_data():
    row_cnt = 1000
    header = ["int_id", "str_id", "int_random", "bool_random", "float_random"]

    int_id = list(range(row_cnt))
    str_id = list("id_" + str(i) for i in range(row_cnt))

    int_random = np.random.randint(100, size=row_cnt)
    bool_random = int_random < 50
    float_random = np.random.randn(row_cnt)

    X = [
        [int_id[i], str_id[i], int_random[i], bool_random[i], float_random[i]]
        for i in range(row_cnt)
    ]
    # Convert the list of lists to a DataFrame
    df = pd.DataFrame(X, columns=header)
    yield df


def calculate_mean_and_variance(df, numeric_df):
    if not isinstance(numeric_df, list):
        raise ValueError("numeric_df should be a list of column names.")
    for col in numeric_df:
        if col not in df.columns:
            raise ValueError(f"Column {col} does not exist in the DataFrame.")
    stats = {}
    for col in numeric_df:
        mean = df[col].mean()
        variance = df[col].var()
        stats[col] = {"mean": mean, "variance": variance}
    return stats


def test_numeric_transformer_fit_test_df(df_data: pd.DataFrame):
    """ """
    # get metadata
    metadata_df = Metadata.from_dataframe(df_data)

    # fit the transformer
    transformer = NumericValueTransformer()
    transformer.fit(metadata_df, df_data)
    assert transformer.int_columns == {"int_random", "int_id"}
    assert transformer.float_columns == {"float_random"}


def test_numeric_transformer_convert_test_df(df_data: pd.DataFrame):
    """ """
    # get metadata
    metadata_df = Metadata.from_dataframe(df_data)
    # fit the transformer
    transformer = NumericValueTransformer()
    transformer.fit(metadata_df, df_data)
    # convert the data
    converted_df = transformer.convert(df_data)
    numerical_columns = list(transformer.int_columns) + list(transformer.float_columns)
    converted_status = calculate_mean_and_variance(converted_df, numerical_columns)
    assert type(converted_df) == pd.DataFrame
    assert converted_df.shape == df_data.shape
    assert np.isclose(converted_status["int_id"]["mean"], 0.0)
    assert np.isclose(converted_status["int_random"]["mean"], 0.0)
    assert np.isclose(converted_status["float_random"]["mean"], 0.0)
    assert np.isclose(converted_status["int_id"]["variance"], 1, atol=0.001)
    assert np.isclose(converted_status["int_random"]["variance"], 1, atol=0.001)
    assert np.isclose(converted_status["float_random"]["variance"], 1, atol=0.001)


def test_numeric_transformer_reverse_convert_test_df(df_data: pd.DataFrame):
    """ """
    # fit the transformer
    transformer = NumericValueTransformer()
    transformer.fit(Metadata.from_dataframe(df_data), df_data)
    numerical_columns = list(transformer.int_columns) + list(transformer.float_columns)
    # convert the data
    converted_df = transformer.convert(df_data)
    # invert the converted data
    reverse_converted_df = transformer.reverse_convert(converted_df)
    reverse_converted_status = calculate_mean_and_variance(reverse_converted_df, numerical_columns)
    original_status = calculate_mean_and_variance(df_data, numerical_columns)
    assert type(reverse_converted_df) == pd.DataFrame
    assert reverse_converted_df.shape == df_data.shape
    assert np.isclose(reverse_converted_status["int_id"]["mean"], original_status["int_id"]["mean"])
    assert np.isclose(
        reverse_converted_status["int_random"]["mean"], original_status["int_random"]["mean"]
    )
    assert np.isclose(
        reverse_converted_status["float_random"]["mean"], original_status["float_random"]["mean"]
    )
    assert np.isclose(
        reverse_converted_status["int_id"]["variance"], original_status["int_id"]["variance"]
    )
    assert np.isclose(
        reverse_converted_status["int_random"]["variance"],
        original_status["int_random"]["variance"],
    )
    assert np.isclose(
        reverse_converted_status["float_random"]["variance"],
        original_status["float_random"]["variance"],
    )



"""Transformers module."""

import importlib
import inspect
from collections import defaultdict
from copy import deepcopy
from functools import lru_cache

from sdgx.models.components.sdv_rdt.transformers.base import BaseTransformer
from sdgx.models.components.sdv_rdt.transformers.boolean import BinaryEncoder
from sdgx.models.components.sdv_rdt.transformers.categorical import (
    CustomLabelEncoder,
    FrequencyEncoder,
    LabelEncoder,
    NormalizedFrequencyEncoder,
    NormalizedLabelEncoder,
    OneHotEncoder,
)
from sdgx.models.components.sdv_rdt.transformers.datetime import (
    OptimizedTimestampEncoder,
    UnixTimestampEncoder,
)
from sdgx.models.components.sdv_rdt.transformers.null import NullTransformer
from sdgx.models.components.sdv_rdt.transformers.numerical import (
    ClusterBasedNormalizer,
    FloatFormatter,
    GaussianNormalizer,
)
from sdgx.models.components.sdv_rdt.transformers.pii.anonymizer import (
    AnonymizedFaker,
    PseudoAnonymizedFaker,
)
from sdgx.models.components.sdv_rdt.transformers.text import RegexGenerator

__all__ = [
    "BaseTransformer",
    "BinaryEncoder",
    "ClusterBasedNormalizer",
    "CustomLabelEncoder",
    "FloatFormatter",
    "FrequencyEncoder",
    "GaussianNormalizer",
    "LabelEncoder",
    "NullTransformer",
    "OneHotEncoder",
    "NormalizedLabelEncoder",
    "NormalizedFrequencyEncoder",
    "OptimizedTimestampEncoder",
    "UnixTimestampEncoder",
    "RegexGenerator",
    "AnonymizedFaker",
    "PseudoAnonymizedFaker",
    "get_transformer_name",
    "get_transformer_class",
    "get_transformer_instance",
    "get_transformers_by_type",
    "get_default_transformers",
    "get_default_transformer",
]


def get_transformer_name(transformer):
    """Return the fully qualified path of the transformer.

    Args:
        transformer:
            A transformer class.

    Raises:
        ValueError:
            Crashes when the transformer is not passed as a class.

    Returns:
        string:
            The path of the transformer.
    """
    if inspect.isclass(transformer):
        return transformer.__module__ + "." + transformer.__name__

    raise ValueError(f"The transformer {transformer} must be passed as a class.")


TRANSFORMERS = {
    get_transformer_name(transformer): transformer
    for transformer in BaseTransformer.get_subclasses()
}

DEFAULT_TRANSFORMERS = {
    "numerical": FloatFormatter(missing_value_replacement="mean"),
    "categorical": FrequencyEncoder(),
    "boolean": BinaryEncoder(missing_value_replacement="mode"),
    "datetime": UnixTimestampEncoder(missing_value_replacement="mean"),
}


def get_transformer_class(transformer):
    """Return a ``transformer`` class from a ``str``.

    Args:
        transformer (str):
            Python path.

    Returns:
        BaseTransformer:
            BaseTransformer subclass class object.
    """
    if transformer in TRANSFORMERS:
        return TRANSFORMERS[transformer]

    package, name = transformer.rsplit(".", 1)
    return getattr(importlib.import_module(package), name)


def get_transformer_instance(transformer):
    """Load a new instance of a ``Transformer``.

    The ``transformer`` is expected to be the transformers path as a ``string``,
    a transformer instance or a transformer type.

    Args:
        transformer (str or BaseTransformer):
            String with the transformer path or instance of a BaseTransformer subclass.

    Returns:
        BaseTransformer:
            BaseTransformer subclass instance.
    """
    if isinstance(transformer, BaseTransformer):
        return deepcopy(transformer)

    if inspect.isclass(transformer) and issubclass(transformer, BaseTransformer):
        return transformer()

    return get_transformer_class(transformer)()


@lru_cache()
def get_transformers_by_type():
    """Build a ``dict`` mapping sdtypes to valid existing transformers for that sdtype.

    Returns:
        dict:
            Mapping of sdtypes to a list of existing transformers that take that
            sdtype as an input.
    """
    sdtype_transformers = defaultdict(list)
    transformer_classes = BaseTransformer.get_subclasses()
    for transformer in transformer_classes:
        input_sdtype = transformer.get_input_sdtype()
        sdtype_transformers[input_sdtype].append(transformer)

    return sdtype_transformers


@lru_cache()
def get_default_transformers():
    """Build a ``dict`` mapping sdtypes to a default transformer for that sdtype.

    Returns:
        dict:
            Mapping of sdtypes to a transformer.
    """
    transformers_by_type = get_transformers_by_type()
    defaults = deepcopy(DEFAULT_TRANSFORMERS)
    for sdtype, transformers in transformers_by_type.items():
        if sdtype not in defaults:
            defaults[sdtype] = transformers[0]()

    return defaults


@lru_cache()
def get_default_transformer(sdtype):
    """Get default transformer for a sdtype.

    Returns:
        Transformer:
            Default transformer for sdtype.
    """
    default_transformers = get_default_transformers()
    return default_transformers[sdtype]



from __future__ import annotations

import json
import time
from pathlib import Path

import click

from sdgx.cachers.manager import CacherManager
from sdgx.cli.utils import cli_wrapper, torch_run_warpper
from sdgx.data_connectors.manager import DataConnectorManager
from sdgx.data_exporters.manager import DataExporterManager
from sdgx.data_processors.manager import DataProcessorManager
from sdgx.models.manager import ModelManager
from sdgx.synthesizer import Synthesizer
from sdgx.utils import logger


@click.command()
@torch_run_warpper
@click.option(
    "--save_dir",
    type=str,
    required=True,
    default="",
    help="The directory to save the synthesizer",
)
@click.option(
    "--model",
    type=str,
    required=True,
    help="The name of the model.",
)
@click.option(
    "--model_path",
    type=str,
    default=None,
    help="The path of the model to load",
)
@click.option(
    "--model_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the model for initialization",
)
@click.option(
    "--load_dir",
    type=str,
    default=None,
    help="The directory to load the synthesizer, if it is specified, ``model_path`` will be ignored.",
)
@click.option(
    "--metadata_path",
    type=str,
    default=None,
    help="The path of the metadata to load",
)
@click.option(
    "--data_connector",
    type=str,
    default=None,
    help="The name of the data connector to use",
)
@click.option(
    "--data_connector_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the data connector to use",
)
@click.option(
    "--raw_data_loaders_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the raw data loader to use",
)
@click.option(
    "--processed_data_loaders_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the processed data loader to use",
)
@click.option(
    "--data_processors",
    type=str,
    default=None,
    help="[Comma separated list] The name of the data processors to use, e.g. 'processor_x,processor_y'",
)
@click.option(
    "--data_processors_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the data processors to use",
)
@click.option(
    "--inspector_max_chunk",
    type=int,
    default=None,
    help="The max chunk of the inspector to load",
)
@click.option(
    "--metadata_include_inspectors",
    type=str,
    default=None,
    help="[Comma separated list] The name of the inspectors to include, e.g. 'inspector_x,inspector_y'",
)
@click.option(
    "--metadata_exclude_inspectors",
    type=str,
    default=None,
    help="[Comma separated list] The name of the inspectors to exclude, e.g. 'inspector_x,inspector_y'",
)
@click.option(
    "--inspector_init_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the inspector to use",
)
@click.option(
    "--model_fit_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the model fit method",
)
@click.option(
    "--dry_run",
    type=bool,
    default=False,
    help="Only init the synthesizer without fitting and save.",
)
@cli_wrapper
def fit(
    save_dir: str,
    model: str,
    model_path: str | None,
    model_kwargs: str | None,
    load_dir: str | None,
    metadata_path: str | None,
    data_connector: str | None,
    data_connector_kwargs: str | None,
    raw_data_loaders_kwargs: str | None,
    processed_data_loaders_kwargs: str | None,
    data_processors: str | None,
    data_processors_kwargs: str | None,
    # ``fit`` args
    inspector_max_chunk: int | None,
    metadata_include_inspectors: str | None,
    metadata_exclude_inspectors: str | None = None,
    inspector_init_kwargs: str | None = None,
    model_fit_kwargs: str | None = None,
    # Others
    dry_run: bool = False,
):
    """
    Fit the synthesizer or load a synthesizer for fitnuning/retraining/continue training...
    """
    if data_processors is not None:
        data_processors = data_processors.strip().split(",")

    if model_kwargs is not None:
        model_kwargs = json.loads(model_kwargs)
    if data_connector_kwargs is not None:
        data_connector_kwargs = json.loads(data_connector_kwargs)
    if raw_data_loaders_kwargs is not None:
        raw_data_loaders_kwargs = json.loads(raw_data_loaders_kwargs)
    if processed_data_loaders_kwargs is not None:
        processed_data_loaders_kwargs = json.loads(processed_data_loaders_kwargs)
    if data_processors_kwargs is not None:
        data_processors_kwargs = json.loads(data_processors_kwargs)

    fit_kwargs = {}
    if inspector_max_chunk is not None:
        fit_kwargs["inspector_max_chunk"] = inspector_max_chunk
    if metadata_include_inspectors is not None:
        fit_kwargs["metadata_include_inspectors"] = metadata_include_inspectors.strip().split(",")
    if metadata_exclude_inspectors is not None:
        fit_kwargs["metadata_exclude_inspectors"] = metadata_exclude_inspectors.strip().split(",")
    if inspector_init_kwargs is not None:
        fit_kwargs["inspector_init_kwargs"] = json.loads(inspector_init_kwargs)
    if model_fit_kwargs is not None:
        fit_kwargs["model_fit_kwargs"] = json.loads(model_fit_kwargs)
    if not save_dir:
        save_dir = Path(f"./sdgx-fit-model-{model}-{time.time()}")
    else:
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(parents=True, exist_ok=True)

    if load_dir:
        if model_path:
            logger.warning(
                "Both ``model_path`` and ``load_dir`` are specified, ``model_path`` will be ignored."
            )
        synthesizer = Synthesizer.load(
            load_dir=load_dir,
            metadata_path=metadata_path,
            data_connector=data_connector,
            data_connector_kwargs=data_connector_kwargs,
            raw_data_loaders_kwargs=raw_data_loaders_kwargs,
            processed_data_loaders_kwargs=processed_data_loaders_kwargs,
            data_processors=data_processors,
            data_processors_kwargs=data_processors_kwargs,
        )
    else:
        if model_kwargs and model_path:
            logger.warning(
                "Both ``model_kwargs`` and ``model_path`` are specified, ``model_kwargs`` will be ignored."
            )
        synthesizer = Synthesizer(
            model=model,
            model_kwargs=model_kwargs,
            model_path=model_path,
            metadata_path=metadata_path,
            data_connector=data_connector,
            data_connector_kwargs=data_connector_kwargs,
            raw_data_loaders_kwargs=raw_data_loaders_kwargs,
            processed_data_loaders_kwargs=processed_data_loaders_kwargs,
            data_processors=data_processors,
            data_processors_kwargs=data_processors_kwargs,
        )

    if dry_run:
        return

    synthesizer.fit(**fit_kwargs)
    save_dir = synthesizer.save(save_dir)

    return save_dir.absolute().as_posix()


@click.command()
@torch_run_warpper
@click.option(
    "--load_dir",
    type=str,
    required=True,
    help="The directory to load the synthesizer.",
)
@click.option(
    "--model",
    type=str,
    required=True,
    help="The name of the model.",
)
@click.option(
    "--raw_data_loaders_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the raw data loaders.",
)
@click.option(
    "--processed_data_loaders_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the processed data loaders.",
)
@click.option(
    "--data_processors",
    type=str,
    default=None,
    help="[Comma separated list] The name of the data processors, e.g. 'data_processor_1,data_processor_2'.",
)
@click.option(
    "--data_processors_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the data processors.",
)
@click.option(
    "--count",
    type=int,
    default=100,
    help="The number of samples to generate.",
)
@click.option(
    "--chunksize",
    type=int,
    default=None,
    help="The size of each chunk. If count is very large, chunksize is recommended.",
)
@click.option(
    "--model_sample_args",
    type=str,
    default=None,
    help="[Json String] The kwargs of the model.sample.",
)
@click.option(
    "--data_exporter",
    type=str,
    default="CsvExporter",
    required=True,
    help="The name of the data exporter.",
)
@click.option(
    "--data_exporter_kwargs",
    type=str,
    default=None,
    help="[Json String] The kwargs of the data exporter.",
)
@click.option(
    "--export_dst",
    type=str,
    default=None,
    help="The destination of the exported data.",
)
@click.option(
    "--dry_run",
    type=bool,
    default=False,
    help="Dry run. Only initialize the synthesizer without sampling.",
)
@cli_wrapper
def sample(
    load_dir: str,
    model: str,
    raw_data_loaders_kwargs: str | None,
    processed_data_loaders_kwargs: str | None,
    data_processors: str | None,
    data_processors_kwargs: str | None,
    # ``sample`` args
    count: int,
    chunksize: int | None,
    model_sample_args: str | None,
    # ``exporter`` args
    data_exporter: str,
    data_exporter_kwargs: str | None,
    export_dst: str | None,
    dry_run: bool,
):
    """
    Load a synthesizer and sample.

    ``load_dir`` should contain model and metadata. Please check :ref:`Synthesizer <Synthesizer>`'s `load` method for more details.
    """
    if data_processors is not None:
        data_processors = data_processors.strip().split(",")

    if raw_data_loaders_kwargs is not None:
        raw_data_loaders_kwargs = json.loads(raw_data_loaders_kwargs)
    if processed_data_loaders_kwargs is not None:
        processed_data_loaders_kwargs = json.loads(processed_data_loaders_kwargs)
    if data_processors_kwargs is not None:
        data_processors_kwargs = json.loads(data_processors_kwargs)

    if model_sample_args is not None:
        model_sample_args = json.loads(model_sample_args)

    if data_exporter_kwargs is not None:
        data_exporter_kwargs = json.loads(data_exporter_kwargs)
    else:
        data_exporter_kwargs = {}

    if not export_dst:
        # Assume export to current directory
        export_dst = Path(f"./sdgx-{model}-{time.time()}/sample-data.csv").expanduser().resolve()

    synthesizer = Synthesizer.load(
        load_dir=load_dir,
        model=model,
        raw_data_loaders_kwargs=raw_data_loaders_kwargs,
        processed_data_loaders_kwargs=processed_data_loaders_kwargs,
        data_processors=data_processors,
        data_processors_kwargs=data_processors_kwargs,
    )

    exporter = DataExporterManager().init_exporter(data_exporter, **data_exporter_kwargs)

    if dry_run:
        return

    exporter.write(
        export_dst,
        synthesizer.sample(
            count=count,
            chunksize=chunksize,
            model_sample_args=model_sample_args,
        ),
    )

    return export_dst


@click.command()
@cli_wrapper
def list_models():
    for model_name, model_cls in ModelManager().registed_models.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.command()
@cli_wrapper
def list_data_connectors():
    for (
        model_name,
        model_cls,
    ) in DataConnectorManager().registed_data_connectors.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.command()
@cli_wrapper
def list_data_processors():
    for (
        model_name,
        model_cls,
    ) in DataProcessorManager().registed_data_processors.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.command()
@cli_wrapper
def list_cachers():
    for model_name, model_cls in CacherManager().registed_cachers.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.command()
@cli_wrapper
def list_data_exporters():
    for model_name, model_cls in DataExporterManager().registed_exporters.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.group()
def cli():
    pass


cli.add_command(fit)
cli.add_command(sample)
cli.add_command(list_models)
cli.add_command(list_data_connectors)
cli.add_command(list_data_processors)
cli.add_command(list_cachers)
cli.add_command(list_data_exporters)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import os
from functools import cached_property
from typing import Callable, Generator

import pandas as pd

from sdgx.data_connectors.base import DataConnector


class DataFrameConnector(DataConnector):
    """
    Directly Wraps DataFrame into :ref:`DataConnector`, for small dataset can be loaded all in memory.

    Args:
        df (pd.DataFrame): DataFrame to be wrapped.

    Example:

        .. code-block:: python
            from sdgx.data_connectors.dataframe_connector import DataFrameConnector
            connector = DataFrameConnector(
                df=pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]}),
            )
            df = connector.read()


    """

    def __init__(
        self,
        df: pd.DataFrame,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.df: pd.DataFrame = df

    def _read(self, offset: int = 0, limit: int | None = None) -> pd.DataFrame | None:
        length = self.df.shape[0]
        if offset >= length:
            return None
        limit = limit or length
        return self.df.iloc[offset : min(offset + limit, length)]

    def _columns(self) -> list[str]:
        return list(self.df.columns)

    def _iter(self, offset=0, chunksize=0) -> Generator[pd.DataFrame, None, None]:
        def generator() -> Generator[pd.DataFrame, None, None]:
            length = self.df.shape[0]
            if offset < length:
                current = offset
                while current < length:
                    yield self.df.iloc[current : min(current + chunksize, length)]
                    current += chunksize

        return generator()


from sdgx.data_connectors.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("DataFrameConnector", DataFrameConnector)



from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.stats import entropy, gaussian_kde

from sdgx.metrics.column.base import ColumnMetric


class JSD(ColumnMetric):
    """JSD : Jensen Shannon Divergence

    This class is used to calculate the Jensen Shannon divergence value betweenthe target columns of real data and synthetic data.

    Currently, we support discrete and continuous columns as inputs.
    """

    def __init__(self) -> None:
        super().__init__()
        self.lower_bound = 0
        self.upper_bound = 1
        self.metric_name = "jensen_shannon_divergence"

    @classmethod
    def calculate(
        cls,
        real_data: pd.DataFrame,
        synthetic_data: pd.DataFrame,
        cols: list[str] | None,
        discrete: bool = True,
    ) -> pd.DataFrame:
        """
        Calculate the JSD value between a real column and a synthetic column.

        Args:
            real_data (pd.DataFrame): The real data.

            synthetic_data (pd.DataFrame): The synthetic data.

            cols (list[str]): The target column to calculat JSD metric.

            discrete (bool): Whether this column is a discrete column.

        Returns:
            JSD_val (float): The meteic value.
        """
        if discrete:
            # 对离散变量求
            JSD.check_input(real_data, synthetic_data)
            joint_pd_real = real_data.groupby(cols, dropna=False).size() / len(real_data)
            joint_pd_syn = synthetic_data.groupby(cols, dropna=False).size() / len(synthetic_data)

            # 对齐操作
            joint_pdf_values_real, joint_pdf_values_syn = joint_pd_real.align(
                joint_pd_syn, fill_value=0
            )
        else:
            # 对连续列
            # 一个非常大的问题在于求联合概率密度的数组是N^d问题，所以一旦选取3列以上求联合概率密度时间复杂度就不可接受的高，哪怕只取每个值范围只取100个点都算不完
            # 离散列由于是直接用原始数据进行排列求密度，只涉及一次除法，不管多少列都算的很快
            real_data_T = real_data[cols].values.T  # 转置
            syn_data_T = synthetic_data[cols].values.T

            # 对连续列估计KDE概率密度
            kde_joint_real = gaussian_kde(real_data_T)
            kde_joint_syn = gaussian_kde(syn_data_T)

            # 均匀取点，取值范围选取真实数据集的最大最小范围
            variables_range = [np.linspace(min(col), max(col), 100) for col in real_data_T]
            grid_points = np.meshgrid(*variables_range)
            grid_points_flat = np.vstack([item.ravel() for item in grid_points])

            # 计算概率分布数组
            joint_pdf_values_real = (
                kde_joint_real(grid_points_flat).reshape(grid_points[0].shape).ravel()
            )
            joint_pdf_values_syn = (
                kde_joint_syn(grid_points_flat).reshape(grid_points[0].shape).ravel()
            )

        # 传入概率分布数组
        JSD_val = JSD.jensen_shannon_divergence(joint_pdf_values_real, joint_pdf_values_syn)

        JSD.check_output(JSD_val)

        return JSD_val

    @classmethod
    def check_output(cls, raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the JSD metric.
        """
        instance = cls()
        if raw_metric_value < instance.lower_bound or raw_metric_value > instance.upper_bound:
            raise ValueError

    @classmethod
    def jensen_shannon_divergence(cls, p: float, q: float):
        """Calculate the jensen_shannon_divergence of p and q.

        Args:
            p (float): the input parameter p.

            q (float): the input parameter q.
        """
        # Calculate the average distribution of p and q
        m = 0.5 * (p + q)

        # Calculate KL divergence
        kl_p = entropy(p, m, base=2)
        kl_q = entropy(q, m, base=2)

        # Calculate Jensen Shannon divergence
        js_divergence = 0.5 * (kl_p + kl_q)

        return js_divergence



from __future__ import annotations

import pandas as pd

from sdgx.utils import logger


class ColumnMetric(object):
    """ColumnMetric

    Metrics used to evaluate the quality of synthetic data columns.
    """

    upper_bound = None
    lower_bound = None
    metric_name = "Accuracy"

    def __init__(self) -> None:
        pass

    @classmethod
    def check_input(
        cls, real_data: pd.Series | pd.DataFrame, synthetic_data: pd.Series | pd.DataFrame
    ):
        """Input check for column or table input.

        Args:
            real_data(pd.DataFrame or pd.Series): the real (original) data table / column.

            synthetic_data(pd.DataFrame or pd.Series): the synthetic (generated) data table / column.
        """

        # Input parameter must not contain None value
        if real_data is None or synthetic_data is None:
            raise TypeError("Input contains None.")

        # The data type should be same
        if type(real_data) is not type(synthetic_data):
            raise TypeError("Data type of real_data and synthetic data should be the same.")

        # Check some data-types that must not be allowed
        if type(real_data) in [int, float, str]:
            raise TypeError("real_data's type must not be None, int, float or str")

        # if type is pd.Series, return directly
        if isinstance(real_data, pd.Series) or isinstance(real_data, pd.DataFrame):
            return real_data, synthetic_data

        # if type is not pd.Series or pd.DataFrame tranfer it to Series
        try:
            real_data = pd.Series(real_data)
            synthetic_data = pd.Series(synthetic_data)
            return real_data, synthetic_data
        except Exception as e:
            logger.error(f"An error occurred while converting to pd.Series: {e}")

        return None, None

    @classmethod
    def calculate(
        cls, real_data: pd.Series | pd.DataFrame, synthetic_data: pd.Series | pd.DataFrame
    ):
        """Calculate the metric value between columns between real table and synthetic table.
        Args:
            real_data(pd.DataFrame or pd.Series): the real (original) data table / column.
            synthetic_data(pd.DataFrame or pd.Series): the synthetic (generated) data table / column.
        """
        # This method should first check the input
        # such as:
        real_data, synthetic_data = ColumnMetric.check_input(real_data, synthetic_data)

        raise NotImplementedError()

    @classmethod
    def check_output(cls, raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the JSD metric.
        """
        raise NotImplementedError()

    pass



# Refer CTGAN Version 0.6.0: https://github.com/sdv-dev/CTGAN@a40570e321cb46d798a823f350e1010a0270d804
# Which is Lincensed by MIT License

import os

import numpy as np
import torch

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.models.base import SynthesizerModel


class StatisticSynthesizerModel(SynthesizerModel):
    random_states = None

    def __init__(self, transformer=None, sampler=None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._generator = None
        self.model = None
        self.status = "UNFINED"
        self.model_type = "MODEL_TYPE_UNDEFINED"
        # self.epochs = epochs
        self._device = "CPU"

    def fit(self, metadata: Metadata, dataloader: DataLoader, *args, **kwargs):
        raise NotImplementedError

    def set_device(self, device):
        """Set the `device` to be used ('GPU' or 'CPU')."""
        self._device = device
        if self._generator is not None:
            self._generator.to(self._device)

    def __getstate__(self):
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        state = self.__dict__.copy()
        self.set_device(device_backup)
        if (
            isinstance(self.random_states, tuple)
            and isinstance(self.random_states[0], np.random.RandomState)
            and isinstance(self.random_states[1], torch.Generator)
        ):
            state["_numpy_random_state"] = self.random_states[0].get_state()
            state["_torch_random_state"] = self.random_states[1].get_state()
            state.pop("random_states")
        return state

    def __getstate__(self):
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        state = self.__dict__.copy()
        self.set_device(device_backup)

        random_states = self.random_states
        if (
            isinstance(random_states, tuple)
            and isinstance(random_states[0], np.random.RandomState)
            and isinstance(random_states[1], torch.Generator)
        ):
            state["_numpy_random_state"] = random_states[0].get_state()
            state["_torch_random_state"] = random_states[1].get_state()
            del state["random_states"]

        return state

    def __setstate__(self, state):
        np_state = state.pop("_numpy_random_state", None)
        torch_state = state.pop("_torch_random_state", None)
        if np_state is not None and torch_state is not None:
            current_torch_state = torch.Generator()
            current_torch_state.set_state(torch_state)
            current_numpy_state = np.random.RandomState()
            current_numpy_state.set_state(np_state)
            state["random_states"] = (current_numpy_state, current_torch_state)
        self.__dict__ = state
        # FIXME: Extrace it into a config file
        if not os.getenv("SDG_FORCE_LOAD_CPU"):
            # Prefer cuda if not specified
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            self.set_device(device)

    def save(self, path):
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        torch.save(self, path)
        self.set_device(device_backup)

    @classmethod
    def load(cls, path):
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = torch.load(path)
        model.set_device(device)
        return model

    def set_random_state(self, random_state):
        if random_state is None:
            self.random_states = random_state
        elif isinstance(random_state, int):
            self.random_states = (
                np.random.RandomState(seed=random_state),
                torch.Generator().manual_seed(random_state),
            )
        elif (
            isinstance(random_state, tuple)
            and isinstance(random_state[0], np.random.RandomState)
            and isinstance(random_state[1], torch.Generator)
        ):
            self.random_states = random_state
        else:
            raise TypeError(
                f"`random_state` {random_state} expected to be an int or a tuple of "
                "(`np.random.RandomState`, `torch.Generator`)"
            )



"""Bivariate copulas."""

import numpy as np
import pandas as pd

from sdgx.models.components.sdv_copulas import EPSILON
from sdgx.models.components.sdv_copulas.bivariate.base import Bivariate, CopulaTypes
from sdgx.models.components.sdv_copulas.bivariate.clayton import Clayton
from sdgx.models.components.sdv_copulas.bivariate.frank import Frank
from sdgx.models.components.sdv_copulas.bivariate.gumbel import Gumbel
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix

__all__ = (
    "Bivariate",
    "Clayton",
    "CopulaTypes",
    "Frank",
    "Gumbel",
)


COMPUTE_EMPIRICAL_STEPS = 50


def _compute_empirical(X):
    """Compute empirical distribution.

    Args:
        X(numpy.array): Shape (n,2); Datapoints to compute the empirical(frequentist) copula.

    Return:
        tuple(list):

    """
    z_left = []
    z_right = []
    L = []
    R = []

    U, V = split_matrix(X)
    N = len(U)
    base = np.linspace(EPSILON, 1.0 - EPSILON, COMPUTE_EMPIRICAL_STEPS)
    # See https://github.com/sdv-dev/Copulas/issues/45

    for k in range(COMPUTE_EMPIRICAL_STEPS):
        left = sum(np.logical_and(U <= base[k], V <= base[k])) / N
        right = sum(np.logical_and(U >= base[k], V >= base[k])) / N

        if left > 0:
            z_left.append(base[k])
            L.append(left / base[k] ** 2)

        if right > 0:
            z_right.append(base[k])
            R.append(right / (1 - z_right[k]) ** 2)

    return z_left, L, z_right, R


def _compute_tail(c, z):
    r"""Compute upper concentration function for tail.

    The upper tail concentration function is defined by:

    .. math:: R(z) = \frac{[1 − 2z + C(z, z)]}{(1 − z)^{2}}

    Args:
        c(Iterable): Values of :math:`C(z,z)`.
        z(Iterable): Values for the empirical copula.

    Returns:
        numpy.ndarray

    """
    return (1.0 - 2 * np.asarray(z) + c) / (np.power(1.0 - np.asarray(z), 2))


def _compute_candidates(copulas, left_tail, right_tail):
    """Compute dependencies.

    Args:
        copulas(list[Bivariate]): Fitted instances of bivariate copulas.
        z_left(list):
        z_right(list):

    Returns:
        tuple[list]: Arrays of left and right dependencies for the empirical copula.


    """
    left = []
    right = []

    X_left = np.column_stack((left_tail, left_tail))
    X_right = np.column_stack((right_tail, right_tail))

    for copula in copulas:
        left.append(copula.cumulative_distribution(X_left) / np.power(left_tail, 2))
        right.append(_compute_tail(copula.cumulative_distribution(X_right), right_tail))

    return left, right


def select_copula(X):
    r"""Select best copula function based on likelihood.

    Given out candidate copulas the procedure proposed for selecting the one
    that best fit to a dataset of pairs :math:`\{(u_j, v_j )\}, j=1,2,...n` , is as follows:

    1. Estimate the most likely parameter :math:`\theta` of each copula candidate for the given
       dataset.

    2. Construct :math:`R(z|\theta)`. Calculate the area under the tail for each of the copula
       candidates.

    3. Compare the areas: :math:`a_u` achieved using empirical copula against the ones
       achieved for the copula candidates. Score the outcome of the comparison from 3 (best)
       down to 1 (worst).

    4. Proceed as in steps 2- 3 with the lower tail and function :math:`L`.

    5. Finally the sum of empirical upper and lower tail functions is compared against
       :math:`R + L`. Scores of the three comparisons are summed and the candidate with the
       highest value is selected.

    Args:
        X(np.ndarray): Matrix of shape (n,2).

    Returns:
        copula: Best copula that fits for it.

    """
    frank = Frank()
    frank.fit(X)

    if frank.tau <= 0:
        return frank

    copula_candidates = [frank]

    # append copulas into the candidate list
    for copula_class in [Clayton, Gumbel]:
        try:
            copula = copula_class()
            copula.tau = frank.tau
            copula._compute_theta()
            copula_candidates.append(copula)
        except ValueError:
            pass

    left_tail, empirical_left_aut, right_tail, empirical_right_aut = _compute_empirical(X)
    candidate_left_auts, candidate_right_auts = _compute_candidates(
        copula_candidates, left_tail, right_tail
    )

    empirical_aut = np.concatenate((empirical_left_aut, empirical_right_aut))
    candidate_auts = [
        np.concatenate((left, right))
        for left, right in zip(candidate_left_auts, candidate_right_auts)
    ]

    # compute L2 distance from empirical distribution
    diff_left = [np.sum((empirical_left_aut - left) ** 2) for left in candidate_left_auts]
    diff_right = [np.sum((empirical_right_aut - right) ** 2) for right in candidate_right_auts]
    diff_both = [np.sum((empirical_aut - candidate) ** 2) for candidate in candidate_auts]

    # calcule ranks
    score_left = pd.Series(diff_left).rank(ascending=False)
    score_right = pd.Series(diff_right).rank(ascending=False)
    score_both = pd.Series(diff_both).rank(ascending=False)

    score = score_left + score_right + score_both

    selected_copula = np.argmax(score.to_numpy())
    return copula_candidates[selected_copula]



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class IDInspector(Inspector):
    _inspect_level = 20
    """
    The inspect_level of IDInspector is higher than NumericInspector.

    Often, some column, especially int type id column can also be recognized as numeric types by NumericInspector, causing the column to be marked repeatedly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ID_columns: set[str] = set()

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """Fit the inspector.

        Gets the list of discrete columns from the raw data.

        Args:
            raw_data (pd.DataFrame): Raw data
        """

        self.ID_columns = set()

        df_length = len(raw_data)
        candidate_columns = set(raw_data.select_dtypes(include=["object", "int64"]).columns)

        for each_col_name in candidate_columns:
            target_col = raw_data[each_col_name]
            col_set_length = len(set(target_col))
            if col_set_length == df_length:
                self.ID_columns.add(each_col_name)

        self.ready = True

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""

        return {"id_columns": list(self.ID_columns)}


@hookimpl
def register(manager):
    manager.register("IDInspector", IDInspector)



import pytest

from sdgx.data_connectors.manager import DataConnectorManager


@pytest.fixture
def manager():
    yield DataConnectorManager()


def test_registed_data_connector(manager: DataConnectorManager):
    assert manager._normalize_name("DummyDataConnector") in manager.registed_data_connectors


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from sdgx.utils import logger


class MultiTableMetric:
    """MultiTableMetric

    Metrics used to evaluate the quality of synthetic multi-table data.
    """

    upper_bound = None
    lower_bound = None
    metric_name = None
    metadata = None
    table_list = []

    def __init__(self, metadata: dict) -> None:
        """Initialization

        Args:
            metadata(dict): This parameter accepts a metadata description dict, which is used to describe the table relations and column description information for each table.
        """
        self.metadata = metadata

    @classmethod
    def check_input(cls, real_data: dict, synthetic_data: dict):
        """Format check for single table input.

        The `real_data` and `synthetic_data` should be dict, which contains tables (in pd.DataFrame).

        Args:
            real_data(dict): the real (original) data table.

            synthetic_data(dict): the synthetic (generated) data table.
        """
        if real_data is None or synthetic_data is None:
            raise TypeError("Input contains None.")

        # The data type should be same
        if type(real_data) is not type(synthetic_data):
            raise TypeError("Data type of real_data and synthetic data should be the same.")

        # if type is dict, return directly
        if (
            isinstance(real_data, dict)
            and len(real_data.keys()) > 0
            and len(synthetic_data.keys()) > 0
        ):
            return real_data, synthetic_data

        logger.error("An error occurred while checking the input.")

        return None, None

    # not a class method
    def calculate(self, real_data: dict, synthetic_data: dict):
        """Calculate the metric value between real tables and synthetic tables.

        Args:

            real_data(dict): the real (original) data table.

            synthetic_data(dict): the synthetic (generated) data table.
        """
        raise NotImplementedError()

    @classmethod
    def check_output(raw_metric_value: float):
        """Check the output value.
        Args:

            raw_metric_value (float):  the calculated raw value of the Mutual Information Similarity.
        """
        raise NotImplementedError()

    pass



from __future__ import annotations

import pandas as pd

from sdgx.data_processors.base import DataProcessor


class Formatter(DataProcessor):
    """
    Base class for formatters.

    Formatter is used to convert data column from one format to another.

    For example, parse datetime into timestamp when trainning, then format timestamp into datetime when sampling.

    Difference with :ref:`Transformer`:
    - When a single column is used as input, use formatter for formatting issues.
    - Formatter will not add additional columns to the input table.
    - When a whole table is used as input, use :ref:`Transformer`.
    """



import random

import faker
import numpy as np
import pandas as pd

from sdgx.data_connectors.generator_connector import GeneratorConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.models.components.optimize.sdv_ctgan.data_transformer import DataTransformer
from sdgx.models.components.sdv_rdt.transformers import OneHotEncoder


def preparing_data():
    fake = faker.Faker()
    data = [
        (
            i,
            random.choice("abcdefg"),
            (random.random() - 0.5) * 1000,
            fake.name(),
            fake.date_between(start_date="today", end_date="+1y"),
            (random.random() - 0.5) * 1000,
            fake.sentence(nb_words=3),
        )
        for i in range(1000)
    ]
    df = pd.DataFrame(data, columns=["id", "grade", "num2", "author", "date", "num", "title"])

    def gen_func():
        yield df.copy()

    connector = GeneratorConnector(gen_func)
    data_metadata = Metadata.from_dataframe(df)
    dl = DataLoader(connector)
    data_metadata.datetime_format = {key: "%Y/%m/%d" for key in data_metadata.datetime_columns}
    transformer = DataTransformer()
    transformer.fit(dl, data_metadata.discrete_columns)
    return transformer, dl


def find_not_matching_column_type_onehot(data, column_transform_info_list):
    col_index = 0
    for column_transform_info in column_transform_info_list:
        output_dim = column_transform_info.output_dimensions
        if column_transform_info.column_type == "discrete" and isinstance(
            column_transform_info.transform, OneHotEncoder
        ):
            arr = data[:, col_index : col_index + output_dim]
            # if bug occurred, the arr is switched as continuous
            print(
                f"Filter not one-hot data for column {column_transform_info.column_name}: ",
                arr[(arr != 0) & (arr != 1)],
            )
            assert np.all((arr == 0) | (arr == 1))
        col_index += output_dim


def test_parallel_transform_fixed_not_columns_switching():
    transformer, data_loader = preparing_data()
    ndarry_loader = transformer._parallel_transform(
        data_loader, transformer._column_transform_info_list
    )
    find_not_matching_column_type_onehot(ndarry_loader, transformer._column_transform_info_list)



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.filter.base import Filter
from sdgx.utils import logger


class PositiveNegativeFilter(Filter):
    """
    A data processor for filtering positive and negative values.

    This filter is used to ensure that values in specific columns remain positive or negative.
    During the reverse conversion process, rows that do not meet the expected positivity or
    negativity will be removed.

    Attributes:
        int_columns (set): A set of column names containing integer values.
        float_columns (set): A set of column names containing float values.
        positive_columns (set): A set of column names that should contain positive values.
        negative_columns (set): A set of column names that should contain negative values.
    """

    int_columns: set
    """
    A set of column names that contain integer values.
    """

    float_columns: set
    """
    A set of column names that contain float values.
    """

    positive_columns: set
    """
    A set of column names that are identified as containing positive numeric values.
    """

    negative_columns: set
    """
    A set of column names that are identified as containing negative numeric values.
    """

    def __init__(self):
        self.int_columns = set()
        self.float_columns = set()
        self.positive_columns = set()
        self.negative_columns = set()

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the data filter.
        """
        logger.info("PositiveNegativeFilter Fitted.")

        # record int and float data
        self.int_columns = metadata.int_columns
        self.float_columns = metadata.float_columns

        # record pos and neg
        self.positive_columns = set(metadata.numeric_format["positive"])
        self.negative_columns = set(metadata.numeric_format["negative"])

        self.fitted = True

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method for data filter (No Action).
        """

        logger.info("Converting data using PositiveNegativeFilter... Finished (No Action)")

        return raw_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse_convert method for the pos_neg data filter.

        Iterate through each row of data, check if there are negative values in positive_columns,
        or positive values in negative_columns. If the conditions are not met, discard the row.
        """
        logger.info(
            f"Data reverse-converted by PositiveNegativeFilter Start with Shape: {processed_data.shape}."
        )

        # Create a boolean mask to mark the rows that need to be retained
        mask = pd.Series(True, index=processed_data.index)

        # Check positive_columns
        for col in self.positive_columns:
            if col in processed_data.columns and pd.api.types.is_numeric_dtype(processed_data[col]):
                mask &= processed_data[col] >= 0

        # Check negative_columns
        for col in self.negative_columns:
            if col in processed_data.columns and pd.api.types.is_numeric_dtype(processed_data[col]):
                mask &= processed_data[col] <= 0

        # Apply the mask to filter the data
        filtered_data = processed_data[mask]

        logger.info(
            f"Data reverse-converted by PositiveNegativeFilter with Output Shape: {filtered_data.shape}."
        )

        return filtered_data


@hookimpl
def register(manager):
    manager.register("PositiveNegativeFilter", PositiveNegativeFilter)



"""Clayton module."""

import numpy as np

from sdgx.models.components.sdv_copulas.bivariate.base import Bivariate, CopulaTypes
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix


class Clayton(Bivariate):
    """Class for clayton copula model."""

    copula_type = CopulaTypes.CLAYTON
    theta_interval = [0, float("inf")]
    invalid_thetas = []

    def generator(self, t):
        r"""Compute the generator function for Clayton copula family.

        The generator is a function
        :math:`\psi: [0,1]\times\Theta \rightarrow [0, \infty)`  # noqa: JS101

        that given an Archimedian copula fulfills:
        .. math:: C(u,v) = \psi^{-1}(\psi(u) + \psi(v))

        Args:
            t (numpy.ndarray)

        Returns:
            numpy.ndarray

        """
        self.check_fit()

        return (1.0 / self.theta) * (np.power(t, -self.theta) - 1)

    def probability_density(self, X):
        r"""Compute probability density function for given copula family.

        The probability density(PDF) for the Clayton family of copulas correspond to the formula:

        .. math:: c(U,V) = \frac{\partial^2}{\partial v \partial u}C(u,v) =
            (\theta + 1)(uv)^{-\theta-1}(u^{-\theta} +
            v^{-\theta} - 1)^{-\frac{2\theta + 1}{\theta}}

        Args:
            X (numpy.ndarray)

        Returns:
            numpy.ndarray: Probability density for the input values.

        """
        self.check_fit()

        U, V = split_matrix(X)

        a = (self.theta + 1) * np.power(U * V, -(self.theta + 1))
        b = np.power(U, -self.theta) + np.power(V, -self.theta) - 1
        c = -(2 * self.theta + 1) / self.theta
        return a * np.power(b, c)

    def cumulative_distribution(self, X):
        """Compute the cumulative distribution function for the clayton copula.

        The cumulative density(cdf), or distribution function for the Clayton family of copulas
        correspond to the formula:

        .. math:: C(u,v) = (u^{-θ} + v^{-θ} - 1)^{-1/θ}

        Args:
            X (numpy.ndarray)

        Returns:
            numpy.ndarray: cumulative probability.

        """
        self.check_fit()

        U, V = split_matrix(X)

        if (V == 0).all() or (U == 0).all():
            return np.zeros(V.shape[0])

        else:
            cdfs = [
                (
                    np.power(
                        np.power(U[i], -self.theta) + np.power(V[i], -self.theta) - 1,
                        -1.0 / self.theta,
                    )
                    if (U[i] > 0 and V[i] > 0)
                    else 0
                )
                for i in range(len(U))
            ]

            return np.array(cdfs)

    def percent_point(self, y, V):
        """Compute the inverse of conditional cumulative distribution :math:`C(u|v)^{-1}`.

        Args:
            y (numpy.ndarray): Value of :math:`C(u|v)`.
            v (numpy.ndarray): given value of v.
        """
        self.check_fit()

        if self.theta < 0:
            return V

        else:
            a = np.power(y, self.theta / (-1 - self.theta))
            b = np.power(V, self.theta)

            # If b == 0, self.theta tends to inf,
            # so the next operation tends to 1
            if (b == 0).all():
                return np.ones(len(V))

            return np.power((a + b - 1) / b, -1 / self.theta)

    def partial_derivative(self, X):
        r"""Compute partial derivative of cumulative distribution.

        The partial derivative of the copula(CDF) is the conditional CDF.

        .. math:: F(v|u) = \frac{\partial C(u,v)}{\partial u} =
            u^{- \theta - 1}(u^{-\theta} + v^{-\theta} - 1)^{-\frac{\theta+1}{\theta}}

        Args:
            X (np.ndarray)
            y (float)

        Returns:
            numpy.ndarray: Derivatives

        """
        self.check_fit()

        U, V = split_matrix(X)

        A = np.power(V, -self.theta - 1)

        # If theta tends to inf, A tends to inf
        # And the next partial_derivative tends to 0
        if (A == np.inf).any():
            return np.zeros(len(V))

        B = np.power(V, -self.theta) + np.power(U, -self.theta) - 1
        h = np.power(B, (-1 - self.theta) / self.theta)
        return A * h

    def compute_theta(self):
        r"""Compute theta parameter using Kendall's tau.

        On Clayton copula this is

        .. math:: τ = θ/(θ + 2) \implies θ = 2τ/(1-τ)
        .. math:: θ ∈ (0, ∞)

        On the corner case of :math:`τ = 1`, return infinite.
        """
        if self.tau == 1:
            return np.inf

        return 2 * self.tau / (1 - self.tau)



from enum import Enum, EnumMeta
from functools import cached_property
from typing import Iterable, List

import numpy as np


class StrValuedEnumMeta(EnumMeta):

    def __str__(self):
        return str(self.values)

    __repr__ = __str__

    def __contains__(cls, item):
        if isinstance(item, str):
            return item in cls.values
        elif isinstance(item, Iterable):
            a = np.array(list(item)).astype(str)
            if len(a.shape) == 0 or a.shape[0] == 0:
                return True
            values = set(a)
            return values.issubset(set(cls.values))
        else:
            super().__contains__(item)


class StrValuedBaseEnum(Enum, metaclass=StrValuedEnumMeta):
    def __hash__(self):
        return hash(self.value)

    @property
    def value(self):
        return str(super().value)

    @classmethod
    @property
    def values(cls) -> set:
        if not hasattr(cls, "__VALUES"):
            cls.__VALUES = {i.value for i in cls}
        return cls.__VALUES

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        else:
            return False

    def __str__(self):
        return self.value


def flatten_array(nested, prefix=""):
    """Flatten an array as a dict.

    Args:
        nested (list, numpy.array):
            Iterable to flatten.
        prefix (str):
            Name to append to the array indices. Defaults to ``''``.

    Returns:
        dict:
            Flattened array.
    """
    result = {}
    for index in range(len(nested)):
        prefix_key = "__".join([prefix, str(index)]) if len(prefix) else str(index)

        value = nested[index]
        if isinstance(value, (list, np.ndarray)):
            result.update(flatten_array(value, prefix=prefix_key))

        elif isinstance(value, dict):
            result.update(flatten_dict(value, prefix=prefix_key))

        else:
            result[prefix_key] = value

    return result


IGNORED_DICT_KEYS = ["fitted", "distribution", "type"]


def flatten_dict(nested, prefix=""):
    """Flatten a dictionary.

    This method returns a flatten version of a dictionary, concatenating key names with
    double underscores.

    Args:
        nested (dict):
            Original dictionary to flatten.
        prefix (str):
            Prefix to append to key name. Defaults to ``''``.

    Returns:
        dict:
            Flattened dictionary.
    """
    result = {}

    for key, value in nested.items():
        prefix_key = "__".join([prefix, str(key)]) if len(prefix) else key

        if key in IGNORED_DICT_KEYS and not isinstance(value, (dict, list)):
            continue

        elif isinstance(value, dict):
            result.update(flatten_dict(value, prefix_key))

        elif isinstance(value, (np.ndarray, list)):
            result.update(flatten_array(value, prefix_key))

        else:
            result[prefix_key] = value

    return result


def log_numerical_distributions_error(numerical_distributions, processed_data_columns, logger):
    """Log error when numerical distributions columns don't exist anymore."""
    unseen_columns = numerical_distributions.keys() - set(processed_data_columns)
    for column in unseen_columns:
        logger.info(
            f"Requested distribution '{numerical_distributions[column]}' "
            f"cannot be applied to column '{column}' because it no longer "
            "exists after preprocessing."
        )


def _key_order(key_value):
    parts = []
    for part in key_value[0].split("__"):
        if part.isdigit():
            part = int(part)

        parts.append(part)

    return parts


def unflatten_dict(flat):
    """Transform a flattened dict into its original form.

    Args:
        flat (dict):
            Flattened dict.

    Returns:
        dict:
            Nested dict (if corresponds)
    """
    unflattened = {}

    for key, value in sorted(flat.items(), key=_key_order):
        if "__" in key:
            key, subkey = key.split("__", 1)
            subkey, name = subkey.rsplit("__", 1)

            if name.isdigit():
                column_index = int(name)
                row_index = int(subkey)

                array = unflattened.setdefault(key, [])

                if len(array) == row_index:
                    row = []
                    array.append(row)
                elif len(array) == row_index + 1:
                    row = array[row_index]
                else:
                    # This should never happen
                    raise ValueError("There was an error unflattening the extension.")

                if len(row) == column_index:
                    row.append(value)
                else:
                    # This should never happen
                    raise ValueError("There was an error unflattening the extension.")

            else:
                subdict = unflattened.setdefault(key, {})
                if subkey.isdigit():
                    subkey = int(subkey)

                inner = subdict.setdefault(subkey, {})
                inner[name] = value

        else:
            unflattened[key] = value

    return unflattened


def validate_numerical_distributions(numerical_distributions, metadata_columns):
    """Validate ``numerical_distributions``.

    Raise an error if it's not None or dict, or if its columns are not present in the metadata.

    Args:
        numerical_distributions (dict):
            Dictionary that maps field names from the table that is being modeled with
            the distribution that needs to be used.
        metadata_columns (list):
            Columns present in the metadata.
    """
    if numerical_distributions:
        if not isinstance(numerical_distributions, dict):
            raise TypeError("numerical_distributions can only be None or a dict instance.")

        invalid_columns = numerical_distributions.keys() - set(metadata_columns)
        # if invalid_columns:
        #     raise SynthesizerInputError(
        #         'Invalid column names found in the numerical_distributions dictionary '
        #         f'{invalid_columns}. The column names you provide must be present '
        #         'in the metadata.'
        #     )



import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.transformers.specific_combination import (
    SpecificCombinationTransformer,
)


@pytest.fixture
def train_data():
    return pd.DataFrame(
        {
            "price_usd": [100, 200, 300],
            "price_cny": [700, 1400, 2100],
            "price_eur": [90, 180, 270],
            "size_cm": [10, 20, 30],
            "size_inch": [3.94, 7.87, 11.81],
            "size_m": [0.1, 0.2, 0.3],
        }
    )


@pytest.fixture
def test_data():
    return pd.DataFrame(
        {
            "price_usd": [200, 200, 100],
            "price_cny": [1400, 1400, 2100],
            "price_eur": [90, 270, 270],
            "size_cm": [10, 20, 20],
            "size_inch": [3.94, 7.87, 11.81],
            "size_m": [0.1, 0.3, 0.3],
        }
    )


@pytest.fixture
def expected_data():
    return pd.DataFrame(
        {
            "price_usd": [200, 200, 300],
            "price_cny": [1400, 1400, 2100],
            "price_eur": [180, 180, 270],
            "size_cm": [10, 20, 30],
            "size_inch": [3.94, 7.87, 11.81],
            "size_m": [0.1, 0.2, 0.3],
        }
    )


def test_specific_combination_transformer(train_data, test_data):
    transformer = SpecificCombinationTransformer()
    metadata = Metadata.from_dataframe(train_data)

    combinations = {("price_usd", "price_cny", "price_eur"), ("size_cm", "size_inch", "size_m")}
    metadata.update({"specific_combinations": combinations})

    transformer.fit(metadata=metadata, tabular_data=train_data)
    result = transformer.reverse_convert(test_data)

    # Check if all rows in result adhere to the specified column mapping combinations.
    for cols in combinations:
        result_rows = result[list(cols)].values.tolist()
        train_rows = train_data[list(cols)].values.tolist()
        assert all(
            row in train_rows for row in result_rows
        ), f"Combination {cols} contains some invalid value in original train data."



import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.formatters.int import IntValueFormatter


def int_formatter_df():
    row_cnt = 1000
    header = ["int_id", "str_id", "int_random", "float_random"]

    int_id = list(range(row_cnt))
    str_id = list("id_" + str(i) for i in range(row_cnt))

    int_random = np.random.randint(100, size=row_cnt)
    float_random = np.random.randn(row_cnt)

    X = [[int_id[i], str_id[i], int_random[i], float_random[i]] for i in range(row_cnt)]
    # Convert the list of lists to a DataFrame
    df = pd.DataFrame(X, columns=header)
    return df


def is_an_integer_list(lst):
    """
    Check if all elements in the list are integers or floats that are also integers.

    Parameters:
    lst (list): The list to be checked.

    Returns:
    bool: True if all elements are integers or floats that are also integers, False otherwise.
    """
    return all(isinstance(i, int) or (isinstance(i, float) and i.is_integer()) for i in lst)


@pytest.mark.skip(reason="success in local, failed in GitHub Action")
def test_int_formatter_fit_test_df():
    """
    Test the functionality of the IntValueFormatter class.

    This function tests the following:
    1. The fit method of the IntValueFormatter class.
    2. The addition of a new column to the formatter.
    3. The reverse conversion of the DataFrame.
    4. The checking of integer values in the DataFrame columns.

    Parameters:
    df_data (pd.DataFrame): The DataFrame to be tested.

    Returns:
    None

    Raises:
    AssertionError: If any of the assertions fail.
    """
    df = int_formatter_df()
    # get metadata
    metadata_df = Metadata.from_dataframe(df)

    # fit the formatter
    formatter = IntValueFormatter()
    formatter.fit(metadata_df)
    metadata_df.column_list = ["int_id", "str_id", "int_random", "float_random"]
    assert sorted(metadata_df.column_list) == sorted(
        ["int_id", "str_id", "int_random", "float_random"]
    )
    # We will temporarily comment out this line of code, which runs without issues locally but causes problems in GitHub Actions.
    # It seems that in GitHub Actions, metadata can interfere with each other, resulting in columns that do not exist in the original DataFrame but come from other datasets.
    # We will open another PR to address this issue.
    # assert formatter.int_columns == {"int_random", "int_id"}
    assert "int_random" in formatter.int_columns
    assert "int_id" in formatter.int_columns
    # add float_random column to formatter
    reverse_df = formatter.reverse_convert(df)
    assert is_an_integer_list(reverse_df["int_id"].tolist())
    assert not is_an_integer_list(reverse_df["str_id"].tolist())
    assert is_an_integer_list(reverse_df["int_random"].tolist())



from __future__ import annotations

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class MyOwnInspector(Inspector): ...


@hookimpl
def register(manager):
    manager.register("DummyInspector", MyOwnInspector)



"""Data loading."""

import json

import numpy as np
import pandas as pd


def read_csv(csv_filename, meta_filename=None, header=True, discrete=None):
    """Read a csv file."""
    data = pd.read_csv(csv_filename, header="infer" if header else None)

    if meta_filename:
        with open(meta_filename) as meta_file:
            metadata = json.load(meta_file)

        discrete_columns = [
            column["name"] for column in metadata["columns"] if column["type"] != "continuous"
        ]

    elif discrete:
        discrete_columns = discrete.split(",")
        if not header:
            discrete_columns = [int(i) for i in discrete_columns]

    else:
        discrete_columns = []

    return data, discrete_columns


def read_tsv(data_filename, meta_filename):
    """Read a tsv file."""
    with open(meta_filename) as f:
        column_info = f.readlines()

    column_info_raw = [x.replace("{", " ").replace("}", " ").split() for x in column_info]

    discrete = []
    continuous = []
    column_info = []

    for idx, item in enumerate(column_info_raw):
        if item[0] == "C":
            continuous.append(idx)
            column_info.append((float(item[1]), float(item[2])))
        else:
            assert item[0] == "D"
            discrete.append(idx)
            column_info.append(item[1:])

    meta = {
        "continuous_columns": continuous,
        "discrete_columns": discrete,
        "column_info": column_info,
    }

    with open(data_filename) as f:
        lines = f.readlines()

    data = []
    for row in lines:
        row_raw = row.split()
        row = []
        for idx, col in enumerate(row_raw):
            if idx in continuous:
                row.append(col)
            else:
                assert idx in discrete
                row.append(column_info[idx].index(col))

        data.append(row)

    return np.asarray(data, dtype="float32"), meta["discrete_columns"]


def write_tsv(data, meta, output_filename):
    """Write to a tsv file."""
    with open(output_filename, "w") as f:
        for row in data:
            for idx, col in enumerate(row):
                if idx in meta["continuous_columns"]:
                    print(col, end=" ", file=f)
                else:
                    assert idx in meta["discrete_columns"]
                    print(meta["column_info"][idx][int(col)], end=" ", file=f)

            print(file=f)



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class DiscreteInspector(Inspector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.discrete_columns: set[str] = set()

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """Fit the inspector.

        Gets the list of discrete columns from the raw data.

        Args:
            raw_data (pd.DataFrame): Raw data
        """
        self.discrete_columns = set()

        self.discrete_columns = self.discrete_columns.union(
            set(raw_data.select_dtypes(include="object").columns)
        )
        self.ready = True

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""

        return {"discrete_columns": list(self.discrete_columns)}


@hookimpl
def register(manager):
    manager.register("DiscreteInspector", DiscreteInspector)



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from sdgx.data_connectors.dataframe_connector import DataFrameConnector


@pytest.fixture
def data_for_test():
    return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


def test_dataframe_connector(data_for_test):
    df = data_for_test.copy()
    c = DataFrameConnector(data_for_test)
    assert c._columns() == ["a", "b"]
    assert_frame_equal(c._read(), df)
    assert_frame_equal(c._read(offset=1), df[1:])
    assert_frame_equal(c._read(offset=2), df[2:])
    assert c._read(offset=3) is None

    assert_frame_equal(c._read(offset=0), df)
    assert c._read(offset=5555) is None

    # iter
    for d, g in zip(c.iter(chunksize=3), [data_for_test]):
        assert_frame_equal(d, g)

    for d, g in zip(c.iter(chunksize=2), [data_for_test[:2], data_for_test[2:]]):
        assert_frame_equal(d, g)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest

from sdgx.data_exporters.manager import DataExporterManager


@pytest.fixture
def manager():
    yield DataExporterManager()


@pytest.mark.parametrize(
    "supportd_exporter",
    [
        "CsvExporter",
    ],
)
def test_manager(supportd_exporter, manager: DataExporterManager):
    assert manager._normalize_name(supportd_exporter) in manager.registed_exporters


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.transformers.fixed_combination import (
    FixedCombinationTransformer,
)


@pytest.fixture
def test_fixed_combination_data():
    data = {
        "A": [1, 2, 3, 4, 5],
        "B": [2, 4, 6, 8, 10],  # B is 2 * A
        "C": [5, 5, 5, 5, 5],  # C is constant
        "D": [1, 3, 5, 7, 9],  # D is not related to A or B
        "E": [2, 4, 6, 8, 10],  # E is 2 * A
        "categorical_one": [
            "co1",
            "co3",
            "co2",
            "co9",
            "co1",
        ],  # categorical_one and categorical_two have a one-to-one correspondence.
        "categorical_two": ["ct1", "ct3", "ct2", "ct9", "ct1"],
    }
    df = pd.DataFrame(data)
    yield df


def test_fixed_combination_handling_test_df(test_fixed_combination_data: pd.DataFrame):
    """
    Test the handling of fixed combination columns in a DataFrame.
    This function tests the behavior of a DataFrame when it contains fixed combination columns.
    It is designed to be used in a testing environment, where the DataFrame is passed as an argument.

    Parameters:
    test_fixed_combination_data (pd.DataFrame): The DataFrame to test.

    Returns:
    None

    Raises:
    AssertionError: If the DataFrame does not handle fixed combination columns as expected.
    """

    metadata = Metadata.from_dataframe(test_fixed_combination_data)
    fixed_combinations = metadata.get("fixed_combinations")
    assert fixed_combinations == {
        "A": {"E", "D", "B"},
        "B": {"A", "E", "D"},
        "D": {"A", "E", "B"},
        "E": {"A", "D", "B"},
        "categorical_one": {"categorical_two"},
    }

    # Initialize the FixedCombinationTransformer.
    fixed_combination_transformer = FixedCombinationTransformer()
    # Check if the transformer has not been fitted yet.
    assert fixed_combination_transformer.fitted is False

    # Fit the transformer with the DataFrame.
    fixed_combination_transformer.fit(metadata)

    # Check if the transformer has been fitted after the fit operation.
    assert fixed_combination_transformer.fitted

    # Check the fixed combination columns
    assert fixed_combination_transformer.fixed_combinations == {
        "A": {"E", "D", "B"},
        "B": {"A", "E", "D"},
        "D": {"A", "E", "B"},
        "E": {"A", "D", "B"},
        "categorical_one": {"categorical_two"},
    }

    # Transform the DataFrame using the transformer.
    transformed_df = fixed_combination_transformer.convert(test_fixed_combination_data)

    # Ensure all original columns are retained
    for column in test_fixed_combination_data.columns:
        assert (
            column in transformed_df.columns
        ), f"Column {column} should be retained in the transformed data."

    # Check if the transformed data meets expectations
    assert transformed_df.shape == test_fixed_combination_data.shape


def test_categorical_fixed_combinations(test_fixed_combination_data):
    """Test the fixed combination relationship of categorical variables"""
    # 测试分类变量的固定组合关系
    metadata = Metadata.from_dataframe(test_fixed_combination_data)
    transformer = FixedCombinationTransformer()
    transformer.fit(metadata)

    # Verify that the correspondence between categorical_one and categorical_two is detected
    # 验证categorical_one和categorical_two的对应关系被检测到
    assert "categorical_one" in transformer.fixed_combinations
    assert "categorical_two" in transformer.fixed_combinations["categorical_one"]

    # Verify that the transformed data maintains the original correspondence
    # 验证转换后的数据保持原有的对应关系
    transformed_df = transformer.convert(test_fixed_combination_data)
    assert all(
        transformed_df["categorical_one"].map(
            dict(
                zip(
                    test_fixed_combination_data["categorical_one"],
                    test_fixed_combination_data["categorical_two"],
                )
            )
        )
        == transformed_df["categorical_two"]
    )



from __future__ import annotations

from collections import defaultdict, namedtuple

import pytest

from sdgx.models.statistics.multi_tables.base import MultiTableSynthesizerModel


@pytest.fixture
def demo_base_multi_table_synthesizer(
    demo_multi_table_data_metadata_combiner, demo_multi_table_data_loader
):
    yield MultiTableSynthesizerModel(
        use_dataloader=True,
        metadata_combiner=demo_multi_table_data_metadata_combiner,
        tables_data_loader=demo_multi_table_data_loader,
    )


def test_base_multi_table_synthesizer(demo_base_multi_table_synthesizer):
    KeyTuple = namedtuple("KeyTuple", ["parent", "child"])

    assert demo_base_multi_table_synthesizer.parent_map == defaultdict(None, {"train": "store"})
    assert demo_base_multi_table_synthesizer.child_map == defaultdict(None, {"store": "train"})
    assert demo_base_multi_table_synthesizer._get_all_foreign_keys("train")[0][0] == KeyTuple(
        parent="Store", child="Store"
    )


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import pandas as pd
import pytest

from sdgx.data_processors.base import DataProcessor


@pytest.fixture
def single_demo_data_df(demo_single_table_path):
    df = pd.read_csv(demo_single_table_path)
    return df[["workclass", "age", "occupation"]]


@pytest.fixture
def attach_demo_data_df(demo_single_table_path):
    df = pd.read_csv(demo_single_table_path)
    return df[["education", "capital-gain"]]


@pytest.fixture
def base_data_processor():
    d = DataProcessor()
    yield d


def test_remove_columns(single_demo_data_df: pd.DataFrame, base_data_processor: DataProcessor):
    assert "occupation" in single_demo_data_df.columns
    assert "workclass" in single_demo_data_df.columns
    assert "age" in single_demo_data_df.columns
    result_df = base_data_processor.remove_columns(single_demo_data_df, ["workclass", "occupation"])
    assert "occupation" not in result_df.columns
    assert "workclass" not in result_df.columns
    assert "age" in result_df.columns


def test_attach_columns(
    single_demo_data_df: pd.DataFrame,
    attach_demo_data_df: pd.DataFrame,
    base_data_processor: DataProcessor,
):
    assert "occupation" in single_demo_data_df.columns
    assert "workclass" in single_demo_data_df.columns
    assert "age" in single_demo_data_df.columns
    assert "education" not in single_demo_data_df.columns
    assert "capital-gain" not in single_demo_data_df.columns
    assert "education" in attach_demo_data_df.columns
    assert "capital-gain" in attach_demo_data_df.columns
    attached_df = base_data_processor.attach_columns(single_demo_data_df, attach_demo_data_df)
    assert "occupation" in attached_df.columns
    assert "workclass" in attached_df.columns
    assert "age" in attached_df.columns
    assert "education" in attached_df.columns
    assert "capital-gain" in attached_df



"""Transformers for text data."""

import warnings

import numpy as np

from sdgx.models.components.sdv_rdt.transformers.base import BaseTransformer
from sdgx.models.components.sdv_rdt.transformers.null import NullTransformer
from sdgx.models.components.sdv_rdt.transformers.utils import strings_from_regex


class RegexGenerator(BaseTransformer):
    """RegexGenerator transformer.

    This transformer will drop a column and regenerate it with the previously specified
    ``regex`` format. The transformer will also be able to handle nulls and regenerate null values
    if specified.

    Args:
        regex (str):
            String representing the regex function.
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an integer or float is given,
            replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
            given, replace them with the corresponding aggregation. If ``None`` is given,
            do not replace them. Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
    """

    DETERMINISTIC_TRANSFORM = False
    DETERMINISTIC_REVERSE = False
    INPUT_SDTYPE = "text"
    null_transformer = None

    def __init__(
        self, regex_format="[A-Za-z]{5}", missing_value_replacement=None, model_missing_values=False
    ):
        self.missing_value_replacement = missing_value_replacement
        self.model_missing_values = model_missing_values
        self.regex_format = regex_format
        self.data_length = None

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {}
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes["is_null"] = "float"

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self.null_transformer = NullTransformer(
            self.missing_value_replacement, self.model_missing_values
        )
        self.null_transformer.fit(data)
        self.data_length = len(data)

    def _transform(self, data):
        """Return ``null`` column if ``models_missing_values``.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            (numpy.ndarray or None):
                If ``self.model_missing_values`` is ``True`` then will return a ``numpy.ndarray``
                indicating which values should be ``nan``, else will return ``None``. In both
                scenarios the original column is being dropped.
        """
        if self.null_transformer and self.null_transformer.models_missing_values():
            return self.null_transformer.transform(data)[:, 1].astype(float)

        return None

    def _reverse_transform(self, data):
        """Generate new data using the provided ``regex_format``.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series
        """
        if data is not None and len(data):
            sample_size = len(data)
        else:
            sample_size = self.data_length

        generator, size = strings_from_regex(self.regex_format)
        if sample_size > size:
            warnings.warn(
                f"The data has {sample_size} rows but the regex for '{self.get_input_column()}' "
                f"can only create {size} unique values. Some values in "
                f"'{self.get_input_column()}' may be repeated."
            )

        if size > sample_size:
            reverse_transformed = np.array(
                [next(generator) for _ in range(sample_size)], dtype=object
            )

        else:
            generated_values = list(generator)
            reverse_transformed = []
            while len(reverse_transformed) < sample_size:
                remaining = sample_size - len(reverse_transformed)
                reverse_transformed.extend(generated_values[:remaining])

            reverse_transformed = np.array(reverse_transformed, dtype=object)

        if self.null_transformer.models_missing_values():
            reverse_transformed = np.column_stack((reverse_transformed, data))

        return self.null_transformer.reverse_transform(reverse_transformed)



import shutil

import pytest

from sdgx.models.ml.single_table.ctgan import CTGANSynthesizerModel


@pytest.fixture
def ctgan():
    yield CTGANSynthesizerModel(epochs=1)


@pytest.fixture
def save_model_dir(tmp_path):
    dirname = tmp_path / "model"
    yield dirname
    shutil.rmtree(dirname, ignore_errors=True)


def assert_sampled_data(dummy_single_table_data_loader, sampled_data, count):
    assert len(sampled_data) == count
    assert sampled_data.columns.tolist() == dummy_single_table_data_loader.columns()


def test_ctgan(
    ctgan: CTGANSynthesizerModel,
    dummy_single_table_metadata,
    dummy_single_table_data_loader,
    save_model_dir,
):
    ctgan.fit(dummy_single_table_metadata, dummy_single_table_data_loader)
    sampled_data = ctgan.sample(10)
    assert_sampled_data(dummy_single_table_data_loader, sampled_data, 10)

    ctgan.save(save_model_dir)
    assert save_model_dir.exists()

    model = CTGANSynthesizerModel.load(save_model_dir)
    sampled_data = model.sample(10)
    assert_sampled_data(dummy_single_table_data_loader, sampled_data, 10)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import shutil

import pandas as pd
import pytest

from sdgx.models.components.sdv_rdt.transformers.categorical import (
    NormalizedLabelEncoder,
)


@pytest.fixture(scope="module")
def data_test():
    return pd.DataFrame(
        {
            "x": [str(i) for i in range(100)],
            "y": [str(-i) for i in range(50)] * 2,
            "z": [str(i) for i in range(25)] * 4,
        },
        columns=["x", "y", "z"],
    )


def test_encoder(data_test: pd.DataFrame):

    for col in ["x", "y", "z"]:
        nlabel_encoder = NormalizedLabelEncoder()
        nlabel_encoder.fit(data_test, col)
        td = nlabel_encoder.transform(data_test.copy())
        rd = nlabel_encoder.reverse_transform(td.copy())
        td.rename(columns={f"{col}.value": f"{col}"}, inplace=True)
        assert (rd[col].sort_values().values == data_test[col].sort_values().values).all()
        assert (td[col] >= -1).all()
        assert (td[col] <= 1).all()
        assert td[col].shape == data_test[col].shape
        assert len(td[col].unique()) == len(data_test[col].unique())


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class ColumnOrderTransformer(Transformer):
    """
    A transformer that rearranges the columns of a DataFrame to a specified order.

    Attributes:
        column_list (list): The list of column names in the desired order.

    Methods:
        fit(metadata: Metadata | None = None, **kwargs: dict[str, Any]): Fits the transformer by remembering the order of the columns.
        convert(raw_data: pd.DataFrame) -> pd.DataFrame: Converts the input DataFrame by rearranging its columns.
        reverse_convert(processed_data: pd.DataFrame) -> pd.DataFrame: Reverse-converts the processed DataFrame by rearranging its columns back to their original order.
        rearrange_columns(column_list, processed_data): Rearranges the columns of a DataFrame according to the provided column list.
    """

    column_list: list
    """
    The list of tabular data's columns.
    """

    def __init__(self):
        self.column_list = None

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the transformer.

        Remember the order of the columns.
        """

        self.column_list = list(metadata.column_list)

        logger.info("ColumnOrderTransformer Fitted.")

        self.fitted = True

        return

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method to handle missing values in the input data.
        """
        logger.info("Converting data using ColumnOrderTransformer...")
        logger.info("Converting data using ColumnOrderTransformer... Finished (No action).")

        return raw_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse_convert method for the transformer.
        """

        res = self.rearrange_columns(self.column_list, processed_data)
        logger.info("Data reverse-converted by ColumnOrderTransformer.")

        return res

    @staticmethod
    def rearrange_columns(column_list, processed_data):
        """
        This method rearranges the columns of a given DataFrame according to the provided column list.

        Any columns in the DataFrame that are not in the column list are dropped.

        Args:
            - column_list (list): A list of column names in the order they should appear in the output DataFrame.
            - processed_data (pd.DataFrame): The DataFrame to be rearranged.

        Returns:
            - result_data (pd.DataFrame): The rearranged DataFrame.
        """
        result_data = processed_data.reindex(columns=column_list)

        return result_data


@hookimpl
def register(manager):
    manager.register("ColumnOrderTransformer", ColumnOrderTransformer)



import os

os.environ["SDG_NDARRAY_CACHE_ROOT"] = "/tmp/sdgx/ndarray_cache"

import random
import shutil
import string
from functools import partial

import pandas as pd
import pytest

from sdgx.data_connectors.csv_connector import CsvConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.combiner import MetadataCombiner
from sdgx.data_models.metadata import Metadata
from sdgx.data_models.relationship import Relationship
from sdgx.utils import download_demo_data, download_multi_table_demo_data

_HERE = os.path.dirname(__file__)

# Cache it locally for rerun the tests
DATA_DIR = os.path.join(_HERE, "dataset")


def ramdon_str():
    return "".join(random.choice(string.ascii_letters) for _ in range(10))


@pytest.fixture
def dummy_single_table_path(tmp_path):
    dummy_size = 10
    role_set = ["admin", "user", "guest"]

    df = pd.DataFrame(
        {
            "role": [random.choice(role_set) for _ in range(dummy_size)],
            "name": [ramdon_str() for _ in range(dummy_size)],
            "feature_x": [random.random() for _ in range(dummy_size)],
            "feature_y": [random.random() for _ in range(dummy_size)],
            "feature_z": [random.random() for _ in range(dummy_size)],
        }
    )
    save_path = tmp_path / "dummy.csv"
    df.to_csv(save_path, index=False, header=True)
    yield save_path
    save_path.unlink()


@pytest.fixture
def dummy_single_table_data_connector(dummy_single_table_path):
    yield CsvConnector(
        path=dummy_single_table_path,
    )


@pytest.fixture
def dummy_single_table_data_loader(dummy_single_table_data_connector, cacher_kwargs):
    d = DataLoader(dummy_single_table_data_connector, cacher_kwargs=cacher_kwargs)
    yield d
    d.finalize()


@pytest.fixture
def dummy_single_table_metadata(dummy_single_table_data_loader):
    yield Metadata.from_dataloader(dummy_single_table_data_loader)


@pytest.fixture
def demo_single_table_path():
    yield download_demo_data(DATA_DIR).as_posix()


@pytest.fixture
def demo_relational_table_path(tmp_path):
    dummy_size = 10
    role_set = ["admin", "user", "guest"]

    df = pd.DataFrame(
        {
            "id": list(range(dummy_size)),
            "role": [random.choice(role_set) for _ in range(dummy_size)],
            "name": [ramdon_str() for _ in range(dummy_size)],
            "feature_x": [random.random() for _ in range(dummy_size)],
            "feature_y": [random.random() for _ in range(dummy_size)],
            "feature_z": [random.random() for _ in range(dummy_size)],
        }
    )
    save_path_a = tmp_path / "dummy_relation_A.csv"
    df.to_csv(save_path_a, index=False, header=True)

    sub_size = 5
    assert dummy_size >= sub_size
    df = pd.DataFrame(
        {
            "foreign_id": list(range(sub_size)),
            "feature_i": [random.random() for _ in range(sub_size)],
            "feature_j": [random.random() for _ in range(sub_size)],
            "feature_k": [random.random() for _ in range(sub_size)],
        }
    )
    save_path_b = tmp_path / "dummy_relation_B.csv"
    df.to_csv(save_path_b, index=False, header=True)

    return save_path_a, save_path_b, [("id", "foreign_id")]


@pytest.fixture
def cacher_kwargs(tmp_path):
    cache_dir = tmp_path / "cache"
    yield {"cache_dir": cache_dir.as_posix()}
    shutil.rmtree(cache_dir, ignore_errors=True)


@pytest.fixture
def dataloader_builder(cacher_kwargs):
    yield partial(DataLoader, cacher_kwargs=cacher_kwargs)


@pytest.fixture
def demo_single_table_data_connector(demo_single_table_path):
    yield CsvConnector(
        path=demo_single_table_path,
    )


@pytest.fixture
def demo_single_table_data_loader(demo_single_table_data_connector, cacher_kwargs):
    d = DataLoader(demo_single_table_data_connector, cacher_kwargs=cacher_kwargs)
    yield d
    d.finalize()


@pytest.fixture
def demo_single_table_metadata(demo_single_table_data_loader):
    yield Metadata.from_dataloader(demo_single_table_data_loader)


@pytest.fixture
def demo_multi_table_path():
    yield download_multi_table_demo_data(DATA_DIR)


@pytest.fixture
def demo_multi_table_data_connector(demo_multi_table_path):
    connector_dict = {}
    for each_table in demo_multi_table_path.keys():
        each_path = demo_multi_table_path[each_table]
        connector_dict[each_table] = CsvConnector(path=each_path)
    yield connector_dict


@pytest.fixture
def demo_multi_table_data_loader(demo_multi_table_data_connector, cacher_kwargs):
    loader_dict = {}
    for each_table in demo_multi_table_data_connector.keys():
        each_connector = demo_multi_table_data_connector[each_table]
        each_d = DataLoader(each_connector, cacher_kwargs=cacher_kwargs)
        loader_dict[each_table] = each_d
    yield loader_dict
    for each_table in demo_multi_table_data_connector.keys():
        demo_multi_table_data_connector[each_table].finalize()


@pytest.fixture
def demo_multi_data_parent_matadata(demo_multi_table_data_loader):
    yield Metadata.from_dataloader(demo_multi_table_data_loader["store"])


@pytest.fixture
def demo_multi_data_child_matadata(demo_multi_table_data_loader):
    yield Metadata.from_dataloader(demo_multi_table_data_loader["train"])


@pytest.fixture
def demo_multi_data_relationship():
    yield Relationship.build(parent_table="store", child_table="train", foreign_keys=["Store"])


@pytest.fixture
def demo_multi_table_data_metadata_combiner(
    demo_multi_data_parent_matadata: Metadata,
    demo_multi_data_child_matadata: Metadata,
    demo_multi_data_relationship: Relationship,
):
    # 1. get metadata
    metadata_dict = {}
    metadata_dict["store"] = demo_multi_data_parent_matadata
    metadata_dict["train"] = demo_multi_data_child_matadata
    # 2. define relationship - already defined
    # 3. define combiner
    m = MetadataCombiner(named_metadata=metadata_dict, relationships=[demo_multi_data_relationship])

    yield m



"""Dataset Generators for datetime transformers."""

import datetime
from abc import ABC

import numpy as np
import pandas as pd

from sdgx.models.components.sdv_rdt.performance.datasets.base import (
    BaseDatasetGenerator,
)
from sdgx.models.components.sdv_rdt.performance.datasets.utils import add_nans


class DatetimeGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate datatime data."""

    SDTYPE = "datetime"


class RandomGapDatetimeGenerator(DatetimeGenerator):
    """Generator that creates dates with random gaps between them."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        today = datetime.datetime.today()
        delta = datetime.timedelta(days=1)
        dates = [(np.random.random() * delta + today) for i in range(num_rows)]
        return np.array(dates, dtype="datetime64")

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 300.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }


class RandomGapSecondsDatetimeGenerator(DatetimeGenerator):
    """Generator that creates dates with random gaps of seconds between them."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        today = datetime.datetime.today()
        delta = datetime.timedelta(seconds=1)
        dates = [(np.random.random() * delta + today) for i in range(num_rows)]
        return np.array(dates, dtype="datetime64")

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 300.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }


class RandomGapDatetimeNaNsGenerator(DatetimeGenerator):
    """Generator that creates dates with random gaps and NaNs."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        dates = RandomGapDatetimeGenerator.generate(num_rows)
        return add_nans(dates.astype("O"))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }


class EqualGapHoursDatetimeGenerator(DatetimeGenerator):
    """Generator that creates dates with hour gaps between them."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        today = datetime.datetime.today()
        delta = datetime.timedelta
        dates = [delta(hours=i) + today for i in range(num_rows)]
        return np.array(dates, dtype="datetime64")

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 300.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }


class EqualGapDaysDatetimeGenerator(DatetimeGenerator):
    """Generator that creates dates with 1 day gaps between them."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        today = datetime.datetime.today()
        delta = datetime.timedelta

        today = min(datetime.datetime.today(), pd.Timestamp.max - delta(num_rows))
        dates = [delta(i) + today for i in range(num_rows)]

        return np.array(dates, dtype="datetime64")

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 300.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }


class EqualGapWeeksDatetimeGenerator(DatetimeGenerator):
    """Generator that creates dates with 1 week gaps between them."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        today = datetime.datetime.today()
        delta = datetime.timedelta

        today = datetime.datetime.today()
        dates = [min(delta(weeks=i) + today, pd.Timestamp.max) for i in range(num_rows)]

        return np.array(dates, dtype="datetime64")

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 5e-05, "memory": 500.0},
            "transform": {"time": 5e-05, "memory": 300.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 1000.0,
            },
        }



from __future__ import annotations

import pandas as pd
import pytest

from sdgx.data_connectors.csv_connector import CsvConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.combiner import MetadataCombiner
from sdgx.data_models.inspectors.base import RelationshipInspector
from sdgx.data_models.metadata import Metadata
from sdgx.data_models.relationship import Relationship


class MockInspector(RelationshipInspector):
    def __init__(self, dummy_data: list[Relationship], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tables = set()
        self.dummy_data = dummy_data

    def _build_relationship(self) -> list[Relationship]:
        return self.dummy_data


def test_from_dataloader(demo_relational_table_path, tmp_path):
    table_a_path, table_b_path, pairs = demo_relational_table_path
    dl_a = DataLoader(CsvConnector(path=table_a_path))
    dl_b = DataLoader(CsvConnector(path=table_b_path))
    relationship = Relationship.build(
        parent_table=dl_a.identity,
        parent_metadata=Metadata(primary_keys=["id"], column_list=["id"], id_columns={"id"}),
        child_table=dl_b.identity,
        child_metadata=Metadata(
            primary_keys=["child_id"],
            column_list=["child_id", "foreign_id"],
            id_columns={"child_id", "foreign_id"},
        ),
        foreign_keys=pairs,
    )

    combiner = MetadataCombiner.from_dataloader(
        dataloaders=[dl_a, dl_b],
        metadata_from_dataloader_kwargs={},
        relationshipe_inspector=MockInspector,
        relationships_inspector_kwargs=dict(dummy_data=[relationship]),
    )

    assert dl_a.identity in combiner.named_metadata
    assert dl_b.identity in combiner.named_metadata
    assert combiner.relationships == [relationship]

    save_dir = tmp_path / "unittest-combinner"
    combiner.save(save_dir)
    assert save_dir.exists()
    loaded_combiner = MetadataCombiner.load(save_dir)
    assert combiner == loaded_combiner


def test_from_dataframe(demo_relational_table_path, tmp_path):
    table_a_path, table_b_path, pair = demo_relational_table_path
    relationship = Relationship.build(
        parent_table="table_a",
        parent_metadata=Metadata(primary_keys=["id"], column_list=["id"], id_columns={"id"}),
        child_table="table_b",
        child_metadata=Metadata(
            primary_keys=["child_id"],
            column_list=["child_id", "foreign_id"],
            id_columns={"child_id", "foreign_id"},
        ),
        foreign_keys=pair,
    )

    tb_a = pd.read_csv(table_a_path)
    tb_b = pd.read_csv(table_b_path)

    combiner = MetadataCombiner.from_dataframe(
        dataframes=[tb_a, tb_b],
        names=["table_a", "table_b"],
        metadata_from_dataloader_kwargs={},
        relationshipe_inspector=MockInspector,
        relationships_inspector_kwargs=dict(dummy_data=[relationship]),
    )

    assert "table_a" in combiner.named_metadata
    assert "table_b" in combiner.named_metadata
    assert combiner.relationships == [relationship]

    save_dir = tmp_path / "unittest-combinner"
    combiner.save(save_dir)
    assert save_dir.exists()
    loaded_combiner = MetadataCombiner.load(save_dir)
    assert combiner == loaded_combiner


def test_custom_build_from_dataloaders(demo_relational_table_path, tmp_path):
    table_a_path, table_b_path, pairs = demo_relational_table_path
    dl_a = DataLoader(CsvConnector(path=table_a_path))
    dl_b = DataLoader(CsvConnector(path=table_b_path))
    relationship = Relationship.build(
        parent_table=dl_a.identity,
        parent_metadata=Metadata(primary_keys=["id"], column_list=["id"], id_columns={"id"}),
        child_table=dl_b.identity,
        child_metadata=Metadata(
            primary_keys=["child_id"],
            column_list=["child_id", "foreign_id"],
            id_columns={"child_id", "foreign_id"},
        ),
        foreign_keys=pairs,
    )
    combiner = MetadataCombiner.from_dataloader(
        dataloaders=[dl_a, dl_b],
        metadata_from_dataloader_kwargs={},
        relationshipe_inspector=MockInspector,
        relationships_inspector_kwargs=dict(
            dummy_data=Relationship.build(
                parent_table="balaP",
                parent_metadata=Metadata(
                    primary_keys=["balabala"], column_list=["balabala"], id_columns={"balabala"}
                ),
                child_table="balaC",
                child_metadata=Metadata(
                    primary_keys=["child_id"],
                    column_list=["balabala", "child_id"],
                    id_columns={"balabala", "child_id"},
                ),
                foreign_keys=["balabala"],
            )
        ),
        relationships=[relationship],
    )
    assert dl_a.identity in combiner.named_metadata
    assert dl_b.identity in combiner.named_metadata
    assert combiner.relationships == [relationship]

    save_dir = tmp_path / "unittest-combinner"
    combiner.save(save_dir)
    assert save_dir.exists()
    loaded_combiner = MetadataCombiner.load(save_dir)
    assert combiner == loaded_combiner


def test_custom_build_from_dataframe(demo_relational_table_path, tmp_path):
    table_a_path, table_b_path, pair = demo_relational_table_path
    relationship = Relationship.build(
        parent_table="table_a",
        parent_metadata=Metadata(primary_keys=["id"], column_list=["id"], id_columns={"id"}),
        child_table="table_b",
        child_metadata=Metadata(
            primary_keys=["child_id"],
            column_list=["child_id", "foreign_id"],
            id_columns={"child_id", "foreign_id"},
        ),
        foreign_keys=pair,
    )
    tb_a = pd.read_csv(table_a_path)
    tb_b = pd.read_csv(table_b_path)

    combiner = MetadataCombiner.from_dataframe(
        dataframes=[tb_a, tb_b],
        names=["table_a", "table_b"],
        metadata_from_dataloader_kwargs={},
        relationshipe_inspector=MockInspector,
        relationships_inspector_kwargs=dict(
            dummy_data=Relationship.build(
                parent_table="balaP",
                parent_metadata=Metadata(
                    primary_keys=["balabala"], column_list=["balabala"], id_columns={"balabala"}
                ),
                child_table="balaC",
                child_metadata=Metadata(
                    primary_keys=["child_id"],
                    column_list=["balabala", "child_id"],
                    id_columns={"balabala", "child_id"},
                ),
                foreign_keys=["balabala"],
            )
        ),
        relationships=[relationship],
    )
    assert "table_a" in combiner.named_metadata
    assert "table_b" in combiner.named_metadata
    assert combiner.relationships == [relationship]

    save_dir = tmp_path / "unittest-combinner"
    combiner.save(save_dir)
    assert save_dir.exists()
    loaded_combiner = MetadataCombiner.load(save_dir)
    assert combiner == loaded_combiner


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""BaseTransformer module."""

import abc
import inspect

import pandas as pd


class BaseTransformer:
    """Base class for all transformers.

    The ``BaseTransformer`` class contains methods that must be implemented
    in order to create a new transformer. The ``_fit`` method is optional,
    and ``fit_transform`` method is already implemented.
    """

    INPUT_SDTYPE = None
    SUPPORTED_SDTYPES = None
    OUTPUT_SDTYPES = None
    DETERMINISTIC_TRANSFORM = None
    DETERMINISTIC_REVERSE = None
    COMPOSITION_IS_IDENTITY = None
    NEXT_TRANSFORMERS = None

    columns = None
    column_prefix = None
    output_columns = None

    @classmethod
    def get_subclasses(cls):
        """Recursively find subclasses of this Baseline.

        Returns:
            list:
                List of all subclasses of this class.
        """
        subclasses = []
        for subclass in cls.__subclasses__():
            if abc.ABC not in subclass.__bases__:
                subclasses.append(subclass)

            subclasses += subclass.get_subclasses()

        return subclasses

    @classmethod
    def get_input_sdtype(cls):
        """Return the input sdtype supported by the transformer.

        Returns:
            string:
                Accepted input sdtype of the transformer.
        """
        return cls.INPUT_SDTYPE

    @classmethod
    def get_supported_sdtypes(cls):
        """Return the supported sdtypes by the transformer.

        Returns:
            list:
                Accepted input sdtypes of the transformer.
        """
        return cls.SUPPORTED_SDTYPES or [cls.INPUT_SDTYPE]

    def _add_prefix(self, dictionary):
        if not dictionary:
            return {}

        output = {}
        for output_columns, output_sdtype in dictionary.items():
            output[f"{self.column_prefix}.{output_columns}"] = output_sdtype

        return output

    def get_output_sdtypes(self):
        """Return the output sdtypes produced by this transformer.

        Returns:
            dict:
                Mapping from the transformed column names to the produced sdtypes.
        """
        return self._add_prefix(self.OUTPUT_SDTYPES)

    def is_transform_deterministic(self):
        """Return whether the transform is deterministic.

        Returns:
            bool:
                Whether or not the transform is deterministic.
        """
        return self.DETERMINISTIC_TRANSFORM

    def is_reverse_deterministic(self):
        """Return whether the reverse transform is deterministic.

        Returns:
            bool:
                Whether or not the reverse transform is deterministic.
        """
        return self.DETERMINISTIC_REVERSE

    def is_composition_identity(self):
        """Return whether composition of transform and reverse transform produces the input data.

        Returns:
            bool:
                Whether or not transforming and then reverse transforming returns the input data.
        """
        return self.COMPOSITION_IS_IDENTITY

    def get_next_transformers(self):
        """Return the suggested next transformer to be used for each column.

        Returns:
            dict:
                Mapping from transformed column names to the transformers to apply to each column.
        """
        return self._add_prefix(self.NEXT_TRANSFORMERS)

    def get_input_column(self):
        """Return input column name for transformer.

        Returns:
            str:
                Input column name.
        """
        return self.columns[0]

    def get_output_columns(self):
        """Return list of column names created in ``transform``.

        Returns:
            list:
                Names of columns created during ``transform``.
        """
        return list(self.get_output_sdtypes())

    def _store_columns(self, columns, data):
        if isinstance(columns, tuple) and columns not in data:
            columns = list(columns)
        elif not isinstance(columns, list):
            columns = [columns]

        missing = set(columns) - set(data.columns)
        if missing:
            raise KeyError(f"Columns {missing} were not present in the data.")

        self.columns = columns

    @staticmethod
    def _get_columns_data(data, columns):
        if len(columns) == 1:
            columns = columns[0]

        return data[columns].copy()

    @staticmethod
    def _add_columns_to_data(data, columns, column_names):
        """Add new columns to a ``pandas.DataFrame``.

        Args:
            - data (pd.DataFrame):
                The ``pandas.DataFrame`` to which the new columns have to be added.
            - columns (pd.DataFrame, pd.Series, np.ndarray):
                The data of the new columns to be added.
            - column_names (list, np.ndarray):
                The names of the new columns to be added.

        Returns:
            ``pandas.DataFrame`` with the new columns added.
        """
        if columns is not None:
            if isinstance(columns, (pd.DataFrame, pd.Series)):
                columns.index = data.index

            if len(columns.shape) == 1:
                data[column_names[0]] = columns
            else:
                new_data = pd.DataFrame(columns, columns=column_names)
                data = pd.concat([data, new_data.set_index(data.index)], axis=1)

        return data

    def _build_output_columns(self, data):
        self.column_prefix = "#".join(self.columns)
        self.output_columns = list(self.get_output_sdtypes().keys())

        # make sure none of the generated `output_columns` exists in the data
        data_columns = set(data.columns)
        while data_columns & set(self.output_columns):
            self.column_prefix += "#"
            self.output_columns = list(self.get_output_sdtypes().keys())

    def __repr__(self):
        """Represent initialization of transformer as text.

        Returns:
            str:
                The name of the transformer followed by any non-default parameters.
        """
        class_name = self.__class__.__name__
        custom_args = []
        args = inspect.getfullargspec(self.__init__)
        keys = args.args[1:]
        defaults = args.defaults or []
        defaults = dict(zip(keys, defaults))
        instanced = {key: getattr(self, key) for key in keys}

        if defaults == instanced:
            return f"{class_name}()"

        for arg, value in instanced.items():
            if defaults[arg] != value:
                custom_args.append(f"{arg}={repr(value)}")

        args_string = ", ".join(custom_args)
        return f"{class_name}({args_string})"

    def _fit(self, columns_data):
        """Fit the transformer to the data.

        Args:
            columns_data (pandas.DataFrame or pandas.Series):
                Data to transform.
        """
        raise NotImplementedError()

    def fit(self, data, column):
        """Fit the transformer to a ``column`` of the ``data``.

        Args:
            data (pandas.DataFrame):
                The entire table.
            column (str):
                Column name. Must be present in the data.
        """
        self._store_columns(column, data)

        columns_data = self._get_columns_data(data, self.columns)
        self._fit(columns_data)

        self._build_output_columns(data)

    def _transform(self, columns_data):
        """Transform the data.

        Args:
            columns_data (pandas.DataFrame or pandas.Series):
                Data to transform.

        Returns:
            pandas.DataFrame or pandas.Series:
                Transformed data.
        """
        raise NotImplementedError()

    def transform(self, data, drop=True):
        """Transform the `self.columns` of the `data`.

        Args:
            data (pandas.DataFrame):
                The entire table.
            drop (bool):
                Whether or not to drop original columns.

        Returns:
            pd.DataFrame:
                The entire table, containing the transformed data.
        """
        # if `data` doesn't have the columns that were fitted on, don't transform
        if any(column not in data.columns for column in self.columns):
            return data

        data = data.copy()

        columns_data = self._get_columns_data(data, self.columns)
        transformed_data = self._transform(columns_data)

        data = self._add_columns_to_data(data, transformed_data, self.output_columns)
        if drop:
            data = data.drop(self.columns, axis=1)

        return data

    def fit_transform(self, data, column):
        """Fit the transformer to a `column` of the `data` and then transform it.

        Args:
            data (pandas.DataFrame):
                The entire table.
            column (str):
                A column name.

        Returns:
            pd.DataFrame:
                The entire table, containing the transformed data.
        """
        self.fit(data, column)
        return self.transform(data)

    def _reverse_transform(self, columns_data):
        """Revert the transformations to the original values.

        Args:
            columns_data (pandas.DataFrame or pandas.Series):
                Data to revert.

        Returns:
            pandas.DataFrame or pandas.Series:
                Reverted data.
        """
        raise NotImplementedError()

    def reverse_transform(self, data, drop=True):
        """Revert the transformations to the original values.

        Args:
            data (pandas.DataFrame):
                The entire table.
            drop (bool):
                Whether or not to drop derived columns.

        Returns:
            pandas.DataFrame:
                The entire table, containing the reverted data.
        """
        # if `data` doesn't have the columns that were transformed, don't reverse_transform
        if any(column not in data.columns for column in self.output_columns):
            return data

        data = data.copy()

        columns_data = self._get_columns_data(data, self.output_columns)
        reversed_data = self._reverse_transform(columns_data)

        data = self._add_columns_to_data(data, reversed_data, self.columns)
        if drop:
            data = data.drop(self.output_columns, axis=1)

        return data



from __future__ import annotations

from sdgx.data_processors.base import DataProcessor


class MyOwnDataProcessor(DataProcessor): ...


from sdgx.data_processors.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("DummyDataProcessor", MyOwnDataProcessor)



import copy

import pandas as pd
import pytest

from sdgx.data_models.inspectors.const import ConstInspector


@pytest.fixture
def test_const_data(demo_single_table_path):
    const_col_df = pd.read_csv(demo_single_table_path)

    # Convert the columns to float to allow None values
    const_col_df["age"] = const_col_df["age"].astype(float)
    const_col_df["fnlwgt"] = const_col_df["fnlwgt"].astype(float)

    # Set the values to None
    const_col_df["age"].values[:] = 100
    const_col_df["fnlwgt"].values[:] = 3.14
    const_col_df["workclass"].values[:] = "President"

    yield const_col_df


def test_const_inspector(test_const_data: pd.DataFrame):
    inspector = ConstInspector()
    inspector.fit(test_const_data)
    assert inspector.ready
    assert inspector.const_columns

    assert sorted(inspector.inspect()["const_columns"]) == sorted(["age", "fnlwgt", "workclass"])
    assert inspector.inspect_level == 80


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pandas as pd
import pytest

from sdgx.data_models.inspectors.empty import EmptyInspector


@pytest.fixture
def inspector():
    yield EmptyInspector()


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


@pytest.fixture
def test_empty_data(raw_data: pd.DataFrame):
    # Convert the columns to float to allow None values
    raw_data["age"] = raw_data["age"].astype(float)
    raw_data["fnlwgt"] = raw_data["fnlwgt"].astype(float)

    # Set the values to None
    raw_data["age"].values[:] = None
    raw_data["fnlwgt"].values[:] = None

    yield raw_data


def test_inspector(inspector: EmptyInspector, test_empty_data):
    inspector.fit(test_empty_data)
    assert inspector.ready
    assert inspector.empty_columns
    assert sorted(inspector.inspect()["empty_columns"]) == sorted(
        [
            "age",
            "fnlwgt",
        ]
    )
    assert inspector.inspect_level == 90


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from sdgx.exceptions import SynthesizerInitError
from sdgx.models.base import SynthesizerModel
from sdgx.utils import logger


class LLMBaseModel(SynthesizerModel):
    """
    This is a base class for generating synthetic data using LLM (Large Language Model).

    Note:
    - When using the data loader, the original data is transformed to pd.DataFrame format for subsequent processing.
    - It is not recommended to use this model with large data tables due to excessive token consumption in some expensive LLM service.
    - Generating data based on metadata is a potential way to generate data that cannot be made public and contains sensitive information.
    """

    use_raw_data = False
    """
    By default, we use raw_data for data access.

    When using the data loader, due to the need of randomization operation, we currently use the `.load_all()` to transform the original data to pd.DataFrame format for subsequent processing.

    Due to the characteristics of the OpenAI GPT service, we do not recommend running this model with large data tables, which will consume your tokens excessively.
    """

    use_metadata = False
    """
    In this model, we accept a data generation paradigm that only provides metadata.

    When only metadata is provided, sdgx will format the metadata of the data set into a message and transmit it to GPT, and GPT will generate similar data based on what it knows.

    This is a potential way to generate data that cannot be made public and contains sensitive information.
    """

    _metadata = None
    """
    the metadata.
    """

    off_table_features = []
    """
    * Experimental Feature

    Whether infer data columns that do not exist in the real data table, the effect may not be very good.
    """

    prompts = {
        "message_prefix": """Suppose you are the best data generating model in this world, we have some data samples with the following information:\n\n""",
        "message_suffix": """\nGenerate synthetic data samples based on the above information and your knowledge, each sample should be output on one line (do not output in multiple lines), the output format of the sample is the same as the example in this message, such as "column_name_1 is value_1", the count of the generated data samples is """,
        "system_role_content": "You are a powerful synthetic data generation model.",
    }
    """
    Prompt words for generating data (preliminary version, improvements welcome).
    """

    columns = []
    """
    The columns of the data set.
    """

    dataset_description = ""
    """
    The description of the data set.
    """

    _responses = []
    """
    A list to store the responses received from the LLM.
    """

    _message_list = []
    """
    A list to store the messages used to ask LLM.
    """

    def _check_access_type(self):
        """
        Checks the data access type.

        Raises:
            SynthesizerInitError: If data access type is not specified or if duplicate data access type is found.
        """
        if self.use_dataloader == self.use_raw_data == self.use_metadata == False:
            raise SynthesizerInitError(
                "Data access type not specified, please use `use_raw_data: bool` or `use_dataloader: bool` to specify data access type."
            )
        if self.use_dataloader == self.use_raw_data == True:
            raise SynthesizerInitError("Duplicate data access type found.")

    def _form_columns_description(self):
        """
        We believe that giving information about a column helps improve data quality.

        Currently, we leave this function to Good First Issue until March 2024, if unclaimed we will implement it quickly.
        """

        raise NotImplementedError

    def _form_message_with_offtable_features(self):
        """
        This function forms a message with off-table features.

        If there are more off-table columns, additional processing is excuted here.
        """
        if self.off_table_features:
            logger.info(f"Use off_table_feature = {self.off_table_features}.")
            return f"Also, you should try to infer another {len(self.off_table_features)} columns based on your knowledge, the name of these columns are : {self.off_table_features}, attach these columns after the original table. \n"
        else:
            logger.info("No off_table_feature needed in current model.")
            return ""

    def _form_dataset_description(self):
        """
        This function is used to form the dataset description.

        Returns:
            str: The description of the generated table.
        """
        if self.dataset_description:
            logger.info(f"Use dataset_description = {self.dataset_description}.")
            return "\nThe description of the generated table is " + self.dataset_description + "\n"
        else:
            logger.info("No dataset_description given in current model.")
            return ""



import os

USER_DEFINED_LOG_LEVEL = os.getenv("SDGX_LOG_LEVEL", "INFO")
LOG_TO_FILE = os.getenv("SDGX_LOG_TO_FILE", "false") in ["True", "true"]

os.environ["LOGURU_LEVEL"] = USER_DEFINED_LOG_LEVEL

from loguru import logger


def add_log_file_handler():
    logger.add(
        "sdgx-{time}.log",
        rotation="10 MB",
    )


if LOG_TO_FILE:
    add_log_file_handler()

__all__ = ["logger", "LOG_TO_FILE", "add_log_file_handler", "USER_DEFINED_LOG_LEVEL"]



"""Transformers for categorical data."""

import math
import warnings

import numpy as np
import pandas as pd
import psutil
from scipy.stats import norm

from sdgx.models.components.sdv_rdt.errors import Error
from sdgx.models.components.sdv_rdt.transformers.base import BaseTransformer


class FrequencyEncoder(BaseTransformer):
    """Transformer for categorical data.

    This transformer computes a float representative for each one of the categories
    found in the fit data, and then replaces the instances of these categories with
    the corresponding representative.

    The representatives are decided by sorting the categorical values by their relative
    frequency, then dividing the ``[0, 1]`` interval by these relative frequencies, and
    finally assigning the middle point of each interval to the corresponding category.

    When the transformation is reverted, each value is assigned the category that
    corresponds to the interval it falls in.

    Null values are considered just another category.

    Args:
        add_noise (bool):
            Whether to generate gaussian noise around the class representative of each interval
            or just use the mean for all the replaced values. Defaults to ``False``.
    """

    INPUT_SDTYPE = "categorical"
    SUPPORTED_SDTYPES = ["categorical", "boolean"]
    OUTPUT_SDTYPES = {"value": "float"}
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = True

    mapping = None
    intervals = None
    starts = None
    means = None
    dtype = None

    def __setstate__(self, state):
        """Replace any ``null`` key by the actual ``np.nan`` instance."""
        intervals = state.get("intervals")
        if intervals:
            for key in list(intervals):
                if pd.isna(key):
                    intervals[np.nan] = intervals.pop(key)

        self.__dict__ = state

    def __init__(self, add_noise=False):
        self.add_noise = add_noise

    def is_transform_deterministic(self):
        """Return whether the transform is deterministic.

        Returns:
            bool:
                Whether or not the transform is deterministic.
        """
        return not self.add_noise

    @staticmethod
    def _get_intervals(data, normalized=False):
        """Compute intervals for each categorical value.

        Args:
            data (pandas.Series):
                Data to analyze.

        Returns:
            dict:
                intervals for each categorical value (start, end).
        """
        data = data.fillna(np.nan)
        frequencies = data.value_counts(dropna=False)

        start = -1.0 if normalized else 0.0
        end = 0.0
        elements = len(data)
        probes = frequencies / (elements / 2.0) if normalized else frequencies / elements

        intervals = {}
        means = []
        starts = []
        for value, prob in probes.items():
            end = start + prob
            mean = start + prob / 2
            std = prob / 6
            if pd.isna(value):
                value = np.nan

            intervals[value] = (start, end, mean, std)
            means.append(mean)
            starts.append((value, start))
            start = end

        means = pd.Series(means, index=list(frequencies.keys()))
        starts = pd.DataFrame(starts, columns=["category", "start"]).set_index("start")

        return intervals, means, starts

    def _fit(self, data):
        """Fit the transformer to the data.

        Compute the intervals for each categorical value.

        Args:
            data (pandas.Series):
                Data to fit the transformer to.
        """
        self.dtype = data.dtype
        self.intervals, self.means, self.starts = self._get_intervals(data)

    @staticmethod
    def _clip_noised_transform(result, start, end):
        """Clip transformed values.

        Used to ensure the noise added to transformed values doesn't make it
        go out of the bounds of a given category.

        The upper bound must be slightly lower than ``end``
        so it doesn't get treated as the next category.
        """
        return np.clip(result, start, end - 1e-9)

    def _transform_by_category(self, data):
        """Transform the data by iterating over the different categories."""
        result = np.empty(shape=(len(data),), dtype=float)

        # loop over categories
        for category, values in self.intervals.items():
            start, end, mean, std = values
            if category is np.nan:
                mask = data.isna()
            else:
                mask = data.to_numpy() == category

            if self.add_noise:
                result[mask] = norm.rvs(mean, std, size=mask.sum())
                result[mask] = self._clip_noised_transform(result[mask], start, end)
            else:
                result[mask] = mean

        return result

    def _get_value(self, category):
        """Get the value that represents this category."""
        if pd.isna(category):
            category = np.nan

        start, end, mean, std = self.intervals[category]

        if self.add_noise:
            result = norm.rvs(mean, std)
            return self._clip_noised_transform(result, start, end)

        return mean

    def _transform_by_row(self, data):
        """Transform the data row by row."""
        return data.fillna(np.nan).apply(self._get_value).to_numpy()

    def _transform(self, data):
        """Transform the categorical values to float representatives.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        fit_categories = pd.Series(self.intervals.keys())
        has_nan = pd.isna(fit_categories).any()
        unseen_indexes = ~(data.isin(fit_categories) | (pd.isna(data) & has_nan))
        if unseen_indexes.any():
            # Select only the first 5 unseen categories to avoid flooding the console.
            unseen_categories = set(data[unseen_indexes][:5])
            warnings.warn(
                f"The data contains {unseen_indexes.sum()} new categories that were not "
                f"seen in the original data (examples: {unseen_categories}). Assigning "
                "them random values. If you want to model new categories, "
                "please fit the transformer again with the new data."
            )

        data[unseen_indexes] = np.random.choice(fit_categories, size=unseen_indexes.size)
        if len(self.means) < len(data):
            return self._transform_by_category(data)

        return self._transform_by_row(data)

    def _reverse_transform_by_matrix(self, data):
        """Reverse transform the data with matrix operations."""
        num_rows = len(data)
        num_categories = len(self.starts)

        data = np.broadcast_to(data, (num_categories, num_rows)).T
        starts = np.broadcast_to(self.starts.index, (num_rows, num_categories))
        is_data_greater_than_starts = (data >= starts)[:, ::-1]
        interval_indexes = num_categories - np.argmax(is_data_greater_than_starts, axis=1) - 1

        get_category_from_index = list(self.starts["category"]).__getitem__
        return pd.Series(interval_indexes).apply(get_category_from_index).astype(self.dtype)

    def _reverse_transform_by_category(self, data):
        """Reverse transform the data by iterating over all the categories."""
        result = np.empty(shape=(len(data),), dtype=self.dtype)

        # loop over categories
        for category, values in self.intervals.items():
            start = values[0]
            mask = start <= data.to_numpy()
            result[mask] = category

        return pd.Series(result, index=data.index, dtype=self.dtype)

    def _get_category_from_start(self, value):
        lower = self.starts.loc[:value]
        return lower.iloc[-1].category

    def _reverse_transform_by_row(self, data):
        """Reverse transform the data by iterating over each row."""
        return data.apply(self._get_category_from_start).astype(self.dtype)

    def _reverse_transform(self, data, normalize=False):
        """Convert float values back to the original categorical values.

        Args:
            data (pd.Series):
                Data to revert.

        Returns:
            pandas.Series
        """
        data = data.clip(-1 if normalize else 0, 1)
        num_rows = len(data)
        num_categories = len(self.means)

        # total shape * float size * number of matrices needed
        needed_memory = num_rows * num_categories * 8 * 3
        available_memory = psutil.virtual_memory().available
        if available_memory > needed_memory:
            return self._reverse_transform_by_matrix(data)

        if num_rows > num_categories:
            return self._reverse_transform_by_category(data)

        # loop over rows
        return self._reverse_transform_by_row(data)


class OneHotEncoder(BaseTransformer):
    """OneHotEncoding for categorical data.

    This transformer replaces a single vector with N unique categories in it
    with N vectors which have 1s on the rows where the corresponding category
    is found and 0s on the rest.

    Null values are considered just another category.
    """

    INPUT_SDTYPE = "categorical"
    SUPPORTED_SDTYPES = ["categorical", "boolean"]
    DETERMINISTIC_TRANSFORM = True
    DETERMINISTIC_REVERSE = True

    dummies = None
    _dummy_na = None
    _num_dummies = None
    _dummy_encoded = False
    _indexer = None
    _uniques = None

    @staticmethod
    def _prepare_data(data):
        """Transform data to appropriate format.

        If data is a valid list or a list of lists, transforms it into an np.array,
        otherwise returns it.

        Args:
            data (pandas.Series or pandas.DataFrame):
                Data to prepare.

        Returns:
            pandas.Series or numpy.ndarray
        """
        if isinstance(data, list):
            data = np.array(data)

        if len(data.shape) > 2:
            raise ValueError("Unexpected format.")
        if len(data.shape) == 2:
            if data.shape[1] != 1:
                raise ValueError("Unexpected format.")

            data = data[:, 0]

        return data

    def get_output_sdtypes(self):
        """Return the output sdtypes produced by this transformer.

        Returns:
            dict:
                Mapping from the transformed column names to the produced sdtypes.
        """
        output_sdtypes = {f"value{i}": "float" for i in range(len(self.dummies))}

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Get the pandas `dummies` which will be used later on for OneHotEncoding.

        Args:
            data (pandas.Series or pandas.DataFrame):
                Data to fit the transformer to.
        """
        data = self._prepare_data(data)

        null = pd.isna(data)
        self._uniques = list(pd.unique(data[~null]))
        self._dummy_na = null.any()
        self._num_dummies = len(self._uniques)
        self._indexer = list(range(self._num_dummies))
        self.dummies = self._uniques.copy()

        if not np.issubdtype(data.dtype, np.number):
            self._dummy_encoded = True

        if self._dummy_na:
            self.dummies.append(np.nan)

    def _transform_helper(self, data):
        if self._dummy_encoded:
            coder = self._indexer
            codes = pd.Categorical(data, categories=self._uniques).codes
        else:
            coder = self._uniques
            codes = data

        rows = len(data)
        dummies = np.broadcast_to(coder, (rows, self._num_dummies))
        coded = np.broadcast_to(codes, (self._num_dummies, rows)).T
        array = (coded == dummies).astype(int)

        if self._dummy_na:
            null = np.zeros((rows, 1), dtype=int)
            null[pd.isna(data)] = 1
            array = np.append(array, null, axis=1)

        return array

    def _transform(self, data):
        """Replace each category with the OneHot vectors.

        Args:
            data (pandas.Series, list or list of lists):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        data = self._prepare_data(data)
        unique_data = {np.nan if pd.isna(x) else x for x in pd.unique(data)}
        unseen_categories = unique_data - set(self.dummies)
        if unseen_categories:
            # Select only the first 5 unseen categories to avoid flooding the console.
            examples_unseen_categories = set(list(unseen_categories)[:5])
            warnings.warn(
                f"The data contains {len(unseen_categories)} new categories that were not "
                f"seen in the original data (examples: {examples_unseen_categories}). Creating "
                "a vector of all 0s. If you want to model new categories, "
                "please fit the transformer again with the new data."
            )

        return self._transform_helper(data)

    def _reverse_transform(self, data):
        """Convert float values back to the original categorical values.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to revert.

        Returns:
            pandas.Series
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if data.ndim == 1:
            data = data.reshape(-1, 1)

        indices = np.argmax(data, axis=1)

        return pd.Series(indices).map(self.dummies.__getitem__)


class LabelEncoder(BaseTransformer):
    """LabelEncoding for categorical data.

    This transformer generates a unique integer representation for each category
    and simply replaces each category with its integer value.

    Null values are considered just another category.

    Attributes:
        values_to_categories (dict):
            Dictionary that maps each integer value for its category.
        categories_to_values (dict):
            Dictionary that maps each category with the corresponding
            integer value.

    Args:
        add_noise (bool):
            Whether to generate uniform noise around the label for each category.
            Defaults to ``False``.
        order_by (None or str):
            A string defining how to order the categories before assigning them labels. Defaults to
            ``None``. Options include:
            - ``'numerical_value'``: Order the categories by numerical value.
            - ``'alphabetical'``: Order the categories alphabetically.
            - ``None``: Use the order that the categories appear in when fitting.
    """

    INPUT_SDTYPE = "categorical"
    SUPPORTED_SDTYPES = ["categorical", "boolean"]
    OUTPUT_SDTYPES = {"value": "float"}
    DETERMINISTIC_TRANSFORM = True
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = True

    values_to_categories = None
    categories_to_values = None

    def __init__(self, add_noise=False, order_by=None):
        self.add_noise = add_noise
        if order_by not in [None, "alphabetical", "numerical_value"]:
            raise Error(
                "order_by must be one of the following values: None, 'numerical_value' or "
                "'alphabetical'"
            )

        self.order_by = order_by

    def _order_categories(self, unique_data):
        if self.order_by == "alphabetical":
            if unique_data.dtype.type not in [np.str_, np.object_]:
                raise Error("The data must be of type string if order_by is 'alphabetical'.")

        elif self.order_by == "numerical_value":
            if not np.issubdtype(unique_data.dtype.type, np.number):
                raise Error("The data must be numerical if order_by is 'numerical_value'.")

        if self.order_by is not None:
            nans = pd.isna(unique_data)
            unique_data = np.sort(unique_data[~nans])
            if nans.any():
                unique_data = np.append(unique_data, [np.nan])

        return unique_data

    def _fit(self, data):
        """Fit the transformer to the data.

        Generate a unique integer representation for each category and
        store them in the ``categories_to_values`` dict and its reverse
        ``values_to_categories``.

        Args:
            data (pandas.Series):
                Data to fit the transformer to.
        """
        unique_data = pd.unique(data.fillna(np.nan))
        unique_data = self._order_categories(unique_data)
        self.values_to_categories = dict(enumerate(unique_data))

        self.categories_to_values = {
            category: value for value, category in self.values_to_categories.items()
        }

    def _transform(self, data):
        """Replace each category with its corresponding integer value.

        If a category has not been seen before, a random value is assigned.

        If ``add_noise`` is True, the integer values will be replaced by a
        random number between the value and the value + 1.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            pd.Series
        """
        mapped = data.fillna(np.nan).map(self.categories_to_values)
        is_null = mapped.isna()
        if is_null.any():
            # Select only the first 5 unseen categories to avoid flooding the console.
            unseen_categories = set(data[is_null][:5])
            warnings.warn(
                f"The data contains {is_null.sum()} new categories that were not "
                f"seen in the original data (examples: {unseen_categories}). Assigning "
                "them random values. If you want to model new categories, "
                "please fit the transformer again with the new data."
            )

        mapped[is_null] = np.random.randint(len(self.categories_to_values), size=is_null.sum())

        if self.add_noise:
            mapped = np.random.uniform(mapped, mapped + 1)

        return mapped

    def _reverse_transform(self, data):
        """Convert float values back to the original categorical values.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to revert.

        Returns:
            pandas.Series
        """
        if self.add_noise:
            data = np.floor(data)

        data = data.clip(min(self.values_to_categories), max(self.values_to_categories))
        return data.round().map(self.values_to_categories)


class NormalizedLabelEncoder(LabelEncoder):
    """Same to the LabelEncoder except the transform result will be [-1, 1] instead of positive integer."""

    def __init__(self, order_by=None):
        super().__init__(False, order_by)
        self._round_digit = None

    def _order_categories(self, unique_data):
        # TODO for more check
        if self.order_by == "alphabetical":
            if unique_data.dtype.type not in [np.str_, np.object_]:
                pass

        elif self.order_by == "numerical_value":
            if not np.issubdtype(unique_data.dtype.type, np.number):
                pass

        if self.order_by is not None:
            nans = pd.isna(unique_data)
            unique_data = np.sort(unique_data[~nans])
            if nans.any():
                unique_data = np.append(unique_data, [np.nan])

        return unique_data

    def _fit(self, data):
        """Fit the transformer to the data.

        Generate a unique integer representation for each category and
        store them in the ``categories_to_values`` dict and its reverse
        ``values_to_categories``.

        Args:
            data (pandas.Series):
                Data to fit the transformer to.
        """
        unique_data = pd.unique(data.fillna(np.nan))
        unique_data = self._order_categories(unique_data)

        def normalize_array_to_dict(arr):
            n = len(arr)
            digit = math.ceil(math.log10(n)) + 1
            self._round_digit = digit
            normalized_dict = {round((2 * i / (n - 1)) - 1, digit): arr[i] for i in range(n)}
            return normalized_dict

        self.values_to_categories = normalize_array_to_dict(unique_data)

        self.categories_to_values = {
            category: value for value, category in self.values_to_categories.items()
        }

    def _reverse_transform(self, data):
        """Convert float values back to the original categorical values.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to revert.

        Returns:
            pandas.Series
        """
        value_dict_keys = self.values_to_categories.keys()
        value_dict = self.values_to_categories

        def find_nearest_key(x):
            nearest_key = min(value_dict_keys, key=lambda k: abs(k - x))
            return value_dict[nearest_key]

        data: pd.Series = data.clip(min(self.values_to_categories), max(self.values_to_categories))
        return data.apply(find_nearest_key)


class CustomLabelEncoder(LabelEncoder):
    """Custom label encoder for categorical data.

    This class works very similarly to the ``LabelEncoder``, except that it requires the ordering
    for the labels to be provided.

    Null values are considered just another category.

    Args:
        order (list):
            A list of all the unique categories for the data. The order of the list determines the
            label that each category will get.
        add_noise (bool):
            Whether to generate uniform noise around the label for each category.
            Defaults to ``False``.
    """

    def __init__(self, order, add_noise=False):
        self.order = pd.Series(order).fillna(np.nan)
        super().__init__(add_noise=add_noise)

    def _fit(self, data):
        """Fit the transformer to the data.

        Generate a unique integer representation for each category and
        store them in the ``categories_to_values`` dict and its reverse
        ``values_to_categories``.

        Args:
            data (pandas.Series):
                Data to fit the transformer to.
        """
        data = data.fillna(np.nan)
        missing = list(data[~data.isin(self.order)].unique())
        if len(missing) > 0:
            raise Error(
                f"Unknown categories '{missing}'. All possible categories must be defined in the "
                "'order' parameter."
            )

        self.values_to_categories = dict(enumerate(self.order))
        self.categories_to_values = {
            category: value for value, category in self.values_to_categories.items()
        }


class NormalizedFrequencyEncoder(FrequencyEncoder):
    """Same to FrequencyEncoder except the transform result is in [-1, 1] instead of [0, 1]"""

    def _fit(self, data):
        """Fit the transformer to the data.

        Compute the intervals for each categorical value.

        Args:
            data (pandas.Series):
                Data to fit the transformer to.
        """
        self.dtype = data.dtype
        self.intervals, self.means, self.starts = self._get_intervals(data, normalized=True)

    def _reverse_transform(self, data):
        return super()._reverse_transform(data, True)
